"""
Tests of saving and loading with `pandas_openscm.OpenSCMDB`
"""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest

from pandas_openscm.db import InMemoryDataBackend, InMemoryIndexBackend, OpenSCMDB
from pandas_openscm.testing import create_test_df

filelock = pytest.importorskip("filelock")


def test_lock_is_always_same(tmpdir):
    """
    Test that each instance has its own lock and it is the same on each access

    (This may seem like an odd test, but a previous implementation failed this,
    which led to super weird behaviour.)

    In future, maybe this can be removed,
    but I'd be careful about failing this test.
    If you get the lock, for a given instance,
    it should always be the same thing.
    """
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
    )

    access_one = db.index_file_lock
    access_two = db.index_file_lock

    assert access_one == access_two


def test_lock_acquisition(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
    )

    db.index_file_lock.acquire()
    assert db.index_file_lock.is_locked
    db.index_file_lock.release()
    assert not db.index_file_lock.is_locked


def test_acquire_lock_twice(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
    )

    db.index_file_lock.acquire()
    db.index_file_lock.acquire().lock.acquire(timeout=1.0)

    with db.index_file_lock.acquire():
        db.index_file_lock.acquire(timeout=1.0)


@pytest.mark.parametrize(
    "meth, kwargs",
    (
        ("delete", {}),
        ("load", {}),
        ("load_file_map", {}),
        ("load_metadata", {}),
        (
            "save",
            dict(
                data=create_test_df(
                    variables=(("variable", "kg"),),
                    n_scenarios=1,
                    n_runs=1,
                    timepoints=np.array([1.0, 1.5]),
                ),
                allow_overwrite=True,
            ),
        ),
    ),
)
def test_locking_multi_step(tmpdir, meth, kwargs):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
    )

    # Put some data in the db so there's something to lock
    db.save(
        create_test_df(
            n_scenarios=1,
            variables=[("a", "K")],
            n_runs=1,
            timepoints=np.array([10.0, 15.0]),
        )
    )

    db.index_file_lock.acquire()
    assert db.index_file_lock.is_locked
    # You can acquire the same lock again
    db.index_file_lock.acquire()
    # But if you try to create a new lock,
    # this won't work (either in the same thread of a different one)
    with pytest.raises(filelock.Timeout):
        filelock.FileLock(db.index_file_lock_path).acquire(timeout=0.02)

    # Same logic as above: you can execute methods
    # with the same lock
    getattr(db, meth)(**kwargs)
    with pytest.raises(filelock.Timeout):
        # But a different lock will fail.
        getattr(db, meth)(
            index_file_lock=filelock.FileLock(db.index_file_lock_path, timeout=0.02),
            **kwargs,
        )

    # If we release the lock, now the other paths do work.
    # We have to release the lock twice as we acquired it twice above.
    db.index_file_lock.release()
    db.index_file_lock.release()
    assert not db.index_file_lock.is_locked

    # Can releaes and acquire
    new_lock = filelock.FileLock(db.index_file_lock_path)
    new_lock.acquire(timeout=0.02)
    new_lock.release()
    # Can run methods
    getattr(db, meth)(
        index_file_lock=filelock.FileLock(db.index_file_lock_path, timeout=0.02),
        **kwargs,
    )

    # The same logic applies with context managers
    with db.index_file_lock.acquire():
        assert db.index_file_lock.is_locked
        # Acquiring the same lock again is fine
        db.index_file_lock.acquire()
        # Release again here to avoid double locking
        db.index_file_lock.release()
        # If you try to create a new lock, this won't work.
        with pytest.raises(filelock.Timeout):
            filelock.FileLock(db.index_file_lock_path).acquire(timeout=0.02)

        # You can execute methods that use the instance's lock
        getattr(db, meth)(**kwargs)
        # But a different lock will fail.
        with pytest.raises(filelock.Timeout):
            getattr(db, meth)(
                # Can't use defaults here as default is no timeout
                index_file_lock=(
                    filelock.FileLock(db.index_file_lock_path, timeout=0.02)
                ),
                **kwargs,
            )

    # Out of the context manager, the lock is released
    assert not db.index_file_lock.is_locked
    filelock.FileLock(db.index_file_lock_path).acquire(timeout=0.02)
    getattr(db, meth)(
        # Can't use defaults here as default is no timeout
        index_file_lock=filelock.FileLock(db.index_file_lock_path, timeout=0.02),
        **kwargs,
    )



"""
netCDF backend

Powered by [xarray](https://github.com/pydata/xarray)
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import pandas as pd
from attrs import define

from pandas_openscm.exceptions import MissingOptionalDependencyError

if TYPE_CHECKING:
    import xarray as xr


@define
class netCDFDataBackend:
    """
    netCDF data backend
    """

    ext: str = ".nc"
    """
    Extension to use with files saved by this backend.
    """

    timeseries_dim: str = "ts_id"
    """
    Name of the timeseries dimension in serialised output
    """

    @property
    def preserves_index(self) -> Literal[True]:
        """
        Whether this backend preserves the index of data upon (de-)serialisation
        """
        return True

    def load_data(self, data_file: Path) -> pd.DataFrame:
        """
        Load a data file

        Parameters
        ----------
        data_file
            File from which to load the data

        Returns
        -------
        :
            Loaded data
        """
        try:
            import xarray as xr
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "netCDFBackend.load_data", requirement="xarray"
            ) from exc

        raw = xr.load_dataset(data_file)

        data: pd.DataFrame = raw["values"].to_pandas()  # type: ignore
        index = metadata_xr_to_df(raw)
        index_concat = index.loc[raw[self.timeseries_dim].values]

        res = pd.concat([index_concat, data], axis="columns").set_index(
            index.columns.to_list()
        )

        return res

    def save_data(self, data: pd.DataFrame, data_file: Path) -> None:
        """
        Save data to disk

        Parameters
        ----------
        data
            Data to save

        data_file
            File in which to save the data
        """
        try:
            import xarray as xr
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "netCDFBackend.save_data", requirement="xarray"
            ) from exc

        # Resetting the index will also give each timeseries a unique ID
        data_rs = data.reset_index()
        timeseries_coord_info = {self.timeseries_dim: data_rs.index.values}
        time_coord_info = {"time": data.columns}
        data_index_xr = metadata_df_to_xr(
            data_rs[data.index.names],
            timeseries_id_coord=xr.Coordinates(timeseries_coord_info),
            timeseries_dim=self.timeseries_dim,
        )
        data_values_xr = xr.DataArray(
            data,
            dims=[self.timeseries_dim, "time"],
            coords=xr.Coordinates(timeseries_coord_info | time_coord_info),
        )
        data_xr = xr.merge([data_index_xr, data_values_xr.to_dataset(name="values")])
        data_xr.to_netcdf(data_file)


@define
class netCDFIndexBackend:
    """
    netCDF index backend
    """

    ext: str = ".nc"
    """
    Extension to use with files saved by this backend.
    """

    timeseries_dim: str = "ts_id"
    """
    Name of the timeseries dimension in serialised output
    """

    @property
    def preserves_index(self) -> Literal[True]:
        """
        Whether this backend preserves the `pd.MultiIndex` upon (de-)serialisation
        """
        return True

    @staticmethod
    def load_file_map(file_map_file: Path) -> pd.DataFrame:
        """
        Load the database's file map

        Parameters
        ----------
        file_map_file
            File from which to load the file map

        Returns
        -------
        :
            Loaded file map
        """
        try:
            import xarray as xr
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "netCDFBackend.load_file_map", requirement="xarray"
            ) from exc

        res = xr.load_dataset(file_map_file).to_pandas()
        if isinstance(res, pd.Series):  # pragma: no cover
            raise TypeError(res)

        return res

    @staticmethod
    def load_index(index_file: Path) -> pd.DataFrame:
        """
        Load the index

        Parameters
        ----------
        index_file
            File from which to load the index

        Returns
        -------
        :
            Loaded index
        """
        try:
            import xarray as xr
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "netCDFBackend.load_index", requirement="xarray"
            ) from exc

        raw = xr.load_dataset(index_file)

        intermediate = metadata_xr_to_df(raw)
        res = intermediate.set_index(
            intermediate.columns.difference(["file_id"]).to_list()
        )

        return res

    @staticmethod
    def save_file_map(
        file_map: pd.Series[Path],  # type: ignore # pandas confused about what it supports
        file_map_file: Path,
    ) -> None:
        """
        Save the file map to disk

        Parameters
        ----------
        file_map
            File map to save

        file_map_file
            File in which to save the file map
        """
        try:
            import xarray as xr
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "netCDFBackend.save_file_map", requirement="xarray"
            ) from exc

        file_map_xr = xr.DataArray.from_series(file_map.astype(str))
        file_map_xr.to_netcdf(file_map_file)

    def save_index(
        self,
        index: pd.DataFrame,
        index_file: Path,
    ) -> None:
        """
        Save the index to disk

        Parameters
        ----------
        index
            Index to save

        index_file
            File in which to save the index
        """
        # Use a different name because the timeseries IDs in the index
        # won't necessarily line up with those in the file(s).
        # This should not matter for users, who never see them side-by-side,
        # but just in case.
        index_xr = metadata_df_to_xr(
            # Have to reset the index so we can serialise to disk
            index.reset_index(),
            timeseries_dim=f"{self.timeseries_dim}_index",
        )
        index_xr.to_netcdf(index_file)


def metadata_df_to_xr(
    metadata: pd.DataFrame,
    timeseries_id_coord: xr.Coordinates | None = None,
    timeseries_dim: str = "ts_id",
    map_suffix: str = "_map",
    int_suffix: str = "_int",
) -> xr.Dataset:
    """
    Convert metadata to a format that can be used with xarray

    This assumes that each row is the metadata for an individual timeseries
    and that the index of `metadata` should be used for the unique ID
    of each timeseries.

    This avoids the metadata creating super sparse arrays,
    at the expense of being harder to work with if directly used with xarray.
    If you need to work with super sparse data
    and want to work with it in a lower-dimensional data structure
    to avoid sparsity issues, our advice would be to use pandas.
    For docs on pandas support for higher dimensional data,
    see https://pandas.pydata.org/docs/user_guide/advanced.html#hierarchical-indexing-multiindex.

    Parameters
    ----------
    metadata
        Metadata to convert

    timeseries_id_coord
        Co-ordinate to use for the timeseries ID.

        If not supplied, this will be created from `metadata.index`.

    timeseries_dim
        Name of the dimension to use to label timeseries in the output

    map_suffix
        Suffix to add to metadata names to get their mapping variable

    int_suffix
        Suffix to add to metadata names to get their integer equivalent.

    Returns
    -------
    :
        Metadata, converted to a 'flat' xarray object.

    Raises
    ------
    MissingOptionalDependencyError
        [xarray][] is not installed.

    AssertionError
        `timeseries_id_coord` is `None`
        and the index of `metadata` is not already unique.

    See Also
    --------
    metadata_xr_to_df

    Examples
    --------
    >>> import pandas as pd
    >>>
    >>> metadata = pd.DataFrame(
    ...     [
    ...         ["scen_a", "mod_a", "var_a"],
    ...         ["scen_b", "mod_b", "var_b"],
    ...         ["scen_a", "mod_b", "var_a"],
    ...     ],
    ...     columns=["scenario", "model", "variable"],
    ... )
    >>> metadata_df_to_xr(metadata)  # doctest: +ELLIPSIS
    <xarray.Dataset> Size: ...B
    Dimensions:       (scenario_map: 2, ts_id: 3, model_map: 2, variable_map: 2)
    Coordinates:
      * scenario_map  (scenario_map) int64 16B 0 1
      * ts_id         (ts_id) int64 24B 0 1 2
      * model_map     (model_map) int64 16B 0 1
      * variable_map  (variable_map) int64 16B 0 1
    Data variables:
        scenario      (scenario_map) <U6 48B 'scen_a' 'scen_b'
        scenario_int  (ts_id) int64 24B 0 1 0
        model         (model_map) <U5 40B 'mod_a' 'mod_b'
        model_int     (ts_id) int64 24B 0 1 1
        variable      (variable_map) <U5 40B 'var_a' 'var_b'
        variable_int  (ts_id) int64 24B 0 1 0
    """
    try:
        import xarray as xr
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "metadata_df_to_xr", requirement="xarray"
        ) from exc

    if timeseries_id_coord is None:
        if metadata.index.duplicated().any():
            dups = metadata.index[metadata.index.duplicated(keep=False)]
            msg = f"Index values should be unique. Received these duplicates: {dups}"
            raise AssertionError(msg)

    metadata_columns = metadata.columns.tolist()
    index_darrays = {}
    for metadata_col in metadata_columns:
        metadata_col_unique_vals = metadata[metadata_col].unique().tolist()
        metadata_col_map = {}
        metadata_col_mapped_values = [0] * len(metadata_col_unique_vals)
        for i, v in enumerate(metadata_col_unique_vals):
            metadata_col_map[v] = i
            metadata_col_mapped_values[i] = i

        metadata_map_xr = xr.DataArray(
            metadata_col_unique_vals,
            dims=[f"{metadata_col}{map_suffix}"],
            coords={f"{metadata_col}{map_suffix}": metadata_col_mapped_values},
        )
        metadata_int_xr = xr.DataArray(
            # Not sure if converting to category first then mapping
            # is faster here or not, would have to profile.
            metadata[metadata_col].map(metadata_col_map).astype(int),
            dims=[timeseries_dim],
            coords=timeseries_id_coord,
        )

        index_darrays[metadata_col] = metadata_map_xr
        index_darrays[f"{metadata_col}{int_suffix}"] = metadata_int_xr

    index_xr = xr.Dataset(index_darrays)

    return index_xr


def metadata_xr_to_df(
    metadata: xr.Dataset,
    category_index: bool = False,
    map_suffix: str = "_map",
    int_suffix: str = "_int",
) -> pd.DataFrame:
    """
    Convert metadata in [xarray][] form to [pandas][]

    Parameters
    ----------
    metadata
        Metadata to convert

    category_index
        Should the index be returned as category type?

    map_suffix
        Suffix to add to metadata names to get their mapping variable

    int_suffix
        Suffix to add to metadata names to get their integer equivalent.

    Returns
    -------
    :
        [pandas][] form of the metadata

    See Also
    --------
    metadata_df_to_xr

    Examples
    --------
    >>> import xarray as xr
    >>>
    >>> # You'd almost never write this by hand.
    >>> # However, if you need to, now you see how.
    >>> metadata_xr = xr.Dataset(
    ...     data_vars=dict(
    ...         scenario=xr.DataArray(["scen_a", "scen_b"], dims=("scenario_map",)),
    ...         scenario_int=xr.DataArray([0, 1, 0], dims=("ts_id",)),
    ...         model=xr.DataArray(["mod_a", "mod_b"], dims=("model_map",)),
    ...         model_int=xr.DataArray([0, 1, 1], dims=("ts_id",)),
    ...         variable=xr.DataArray(["var_a", "var_b"], dims=("variable_map",)),
    ...         variable_int=xr.DataArray([0, 1, 0], dims=("ts_id",)),
    ...     ),
    ...     coords=dict(
    ...         scenario_map=[0, 1],
    ...         ts_id=[0, 1, 2],
    ...         model_map=[0, 1],
    ...         variable_map=[0, 1],
    ...     ),
    ... )
    >>>
    >>> res = metadata_xr_to_df(metadata_xr)
    >>> res
          scenario  model variable
    ts_id
    0       scen_a  mod_a    var_a
    1       scen_b  mod_b    var_b
    2       scen_a  mod_b    var_a

    >>> # you can make the result smaller in memory is you use
    >>> # the `category_index` argument.
    >>> metadata_xr_to_df(metadata_xr, category_index=True).dtypes
    scenario    category
    model       category
    variable    category
    dtype: object
    >>>
    >>> # Compared to
    >>> res.dtypes
    scenario    object
    model       object
    variable    object
    dtype: object
    """
    metadata_columns = [
        str(v).replace(map_suffix, "")
        for v in metadata.coords
        if str(v).endswith(map_suffix)
    ]

    index_cols = {}
    for metadata_col in metadata_columns:
        # make naming injectable (?)
        metadata_col_series_int = metadata[f"{metadata_col}{int_suffix}"].to_pandas()
        if category_index:
            metadata_col_series_int = metadata_col_series_int.astype("category")

        metadata_col_map: pd.Series[str] = metadata[metadata_col].to_pandas()  # type: ignore # xarray confused

        index_cols[metadata_col] = metadata_col_series_int.map(metadata_col_map)  # type: ignore # pandas map type hint confused

    res = pd.DataFrame(index_cols)

    return res



"""
Exceptions that are used throughout
"""

from __future__ import annotations


class MissingOptionalDependencyError(ImportError):
    """
    Raised when an optional dependency is missing

    For example, plotting dependencies like matplotlib
    """

    def __init__(self, callable_name: str, requirement: str) -> None:
        """
        Initialise the error

        Parameters
        ----------
        callable_name
            The name of the callable that requires the dependency

        requirement
            The name of the requirement
        """
        error_msg = f"`{callable_name}` requires {requirement} to be installed"
        super().__init__(error_msg)



"""
Tests of `pandas_openscm.grouping` and associated accessors
"""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.grouping import groupby_except
from pandas_openscm.testing import create_test_df


@pytest.mark.parametrize(
    "non_groupers, expected_groups",
    (
        ("run", ["scenario", "variable", "unit"]),
        (["run"], ["scenario", "variable", "unit"]),
        (["run", "scenario"], ["variable", "unit"]),
    ),
)
def test_groupby_except(non_groupers, expected_groups, setup_pandas_accessor):
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "W")),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(10.0),
    )

    pd.testing.assert_frame_equal(
        # Have to do a calculation after the groupby so there is something to compare
        df.groupby(expected_groups).mean(),
        groupby_except(df, non_groupers).mean(),
    )

    # Also test via the accessor
    pd.testing.assert_frame_equal(
        # Have to do a calculation after the groupby so there is something to compare
        df.groupby(expected_groups).mean(),
        df.openscm.groupby_except(non_groupers).mean(),
    )


@pytest.mark.parametrize(
    "new_name, quantile_exp", ((None, "quantile"), ("percentile", "percentile"))
)
def test_fix_index_name_after_groupby_quantile(
    new_name, quantile_exp, setup_pandas_accessor
):
    fix_kwargs = {}
    if new_name is not None:
        fix_kwargs["new_name"] = new_name

    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "W")),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(10.0),
    )

    df_group_raw = df.openscm.groupby_except(["run"]).quantile([0.05, 0.5, 0.95])
    assert df_group_raw.index.names == [
        "scenario",
        "variable",
        "unit",
        None,  # This is the problem we want to fix
    ]

    df_group_fixed = df_group_raw.openscm.fix_index_name_after_groupby_quantile(
        **fix_kwargs
    )
    assert df_group_fixed.index.names == [
        "scenario",
        "variable",
        "unit",
        quantile_exp,  # fixed
    ]



"""
Tests of making plans for moving data with `pandas_openscm.OpenSCMDB`
"""

from __future__ import annotations

import contextlib
from pathlib import Path

import numpy as np
import pandas as pd

from pandas_openscm.db import (
    InMemoryDataBackend,
    InMemoryIndexBackend,
    OpenSCMDB,
)
from pandas_openscm.db.rewriting import MovePlan, ReWriteAction, make_move_plan
from pandas_openscm.testing import assert_move_plan_equal


def test_make_move_plan_no_overwrite(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=contextlib.nullcontext(),  # not used
    )

    index_start = pd.DataFrame(
        [
            ("scenario_a", "variable_a", "Mt", 0),
            ("scenario_a", "variable_b", "Mt", 0),
        ],
        columns=["scenario", "variable", "unit", "file_id"],
    ).set_index(["scenario", "variable", "unit"])
    file_map_start = pd.Series(
        [db.get_new_data_file_path(fid) for fid in index_start["file_id"].unique()],
        index=pd.Index(index_start["file_id"].unique(), name="file_id"),
    )

    index_data_to_write = pd.MultiIndex.from_tuples(
        [
            ("scenario_b", "variable_a", "Mt"),
            ("scenario_b", "variable_b", "Mt"),
        ],
        names=["scenario", "variable", "unit"],
    )
    data_to_write = pd.DataFrame(
        np.random.default_rng().random((index_data_to_write.shape[0], 3)),
        index=index_data_to_write,
        columns=np.arange(2020.0, 2023.0),
    )

    # No overlap so no need to move anything,
    # the index and file map are just the same as what we started with
    # (the layer make_move_plan above deals with writing the new data).
    exp = MovePlan(
        moved_index=index_start,
        moved_file_map=file_map_start,
        rewrite_actions=None,
        delete_paths=None,
    )

    res = make_move_plan(
        index_start=index_start,
        file_map_start=file_map_start,
        data_to_write=data_to_write,
        get_new_data_file_path=db.get_new_data_file_path,
    )

    assert_move_plan_equal(res, exp)


def test_make_move_plan_full_overwrite(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=contextlib.nullcontext(),  # not used
    )

    index_start = pd.DataFrame(
        [
            ("scenario_a", "variable_a", "Mt", 0),
            ("scenario_a", "variable_b", "Mt", 0),
            ("scenario_b", "variable_a", "Mt", 1),
            ("scenario_b", "variable_b", "Mt", 1),
        ],
        columns=["scenario", "variable", "unit", "file_id"],
    ).set_index(["scenario", "variable", "unit"])
    file_map_start = pd.Series(
        [db.get_new_data_file_path(fid) for fid in index_start["file_id"].unique()],
        index=pd.Index(index_start["file_id"].unique(), name="file_id"),
    )

    index_data_to_write = pd.MultiIndex.from_tuples(
        [
            # Full overwrite of file 1
            ("scenario_b", "variable_a", "Mt"),
            ("scenario_b", "variable_b", "Mt"),
        ],
        names=["scenario", "variable", "unit"],
    )
    data_to_write = pd.DataFrame(
        np.random.default_rng().random((index_data_to_write.shape[0], 3)),
        index=index_data_to_write,
        columns=np.arange(2020.0, 2023.0),
    )

    exp_moved_file_ids = [0]  # 1 will be overwritten i.e. schedule to delete
    exp_moved_file_map = pd.Series(
        [db.get_new_data_file_path(file_id) for file_id in exp_moved_file_ids],
        index=pd.Index(exp_moved_file_ids, name="file_id"),
    )

    exp_moved_index = pd.DataFrame(
        [
            # Unchanged
            ("scenario_a", "variable_a", "Mt", 0),
            ("scenario_a", "variable_b", "Mt", 0),
            # # Will be overwritten hence deleted
            # ("scenario_b", "variable_a", "Mt", 1),
            # ("scenario_b", "variable_b", "Mt", 1),
        ],
        columns=["scenario", "variable", "unit", "file_id"],
    ).set_index(index_start.index.names)

    exp = MovePlan(
        moved_index=exp_moved_index,
        moved_file_map=exp_moved_file_map,
        rewrite_actions=None,
        delete_paths=(file_map_start.loc[1],),
    )

    res = make_move_plan(
        index_start=index_start,
        file_map_start=file_map_start,
        data_to_write=data_to_write,
        get_new_data_file_path=db.get_new_data_file_path,
    )

    assert_move_plan_equal(res, exp)


def test_make_move_plan_partial_overwrite(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=contextlib.nullcontext(),  # not used
    )

    index_start = pd.DataFrame(
        [
            ("scenario_a", "variable_a", "Mt", 0),
            ("scenario_a", "variable_b", "Mt", 0),
            ("scenario_b", "variable_a", "Mt", 1),
            ("scenario_b", "variable_b", "Mt", 1),
            ("scenario_c", "variable_a", "Mt", 2),
            ("scenario_c", "variable_b", "Mt", 2),
        ],
        columns=["scenario", "variable", "unit", "file_id"],
    ).set_index(["scenario", "variable", "unit"])
    file_map_start = pd.Series(
        [db.get_new_data_file_path(fid) for fid in index_start["file_id"].unique()],
        index=pd.Index(index_start["file_id"].unique(), name="file_id"),
    )

    index_data_to_write = pd.MultiIndex.from_tuples(
        [
            # File 0 should be left alone
            # ("scenario_a", "variable_a", "Mt"),
            # ("scenario_a", "variable_b", "Mt"),
            # File 1 should be fully deleted to make room
            # for this data
            ("scenario_b", "variable_a", "Mt"),
            ("scenario_b", "variable_b", "Mt"),
            # File 2 should be partially re-written,
            # keeping variable_a but not variable_b
            # (which will be overwritten)
            # ("scenario_c", "variable_a", "Mt"),
            ("scenario_c", "variable_b", "Mt"),
        ],
        names=["scenario", "variable", "unit"],
    )
    data_to_write = pd.DataFrame(
        np.random.default_rng().random((index_data_to_write.shape[0], 3)),
        index=index_data_to_write,
        columns=np.arange(2020.0, 2023.0),
    )

    exp_moved_file_ids = [0, 3]  # 1 deleted, 2 re-written then deleted
    exp_moved_file_map = pd.Series(
        [db.get_new_data_file_path(file_id) for file_id in exp_moved_file_ids],
        index=pd.Index(exp_moved_file_ids, name="file_id"),
    )

    exp_moved_index = pd.DataFrame(
        [
            # Unchanged
            ("scenario_a", "variable_a", "Mt", 0),
            ("scenario_a", "variable_b", "Mt", 0),
            # # Overwritten
            # ("scenario_b", "variable_a", "Mt", 1),
            # ("scenario_b", "variable_b", "Mt", 1),
            # Re-written to make space
            ("scenario_c", "variable_a", "Mt", 3),
            # # Overwritten
            # ("scenario_c", "variable_b", "Mt", 2),
        ],
        columns=["scenario", "variable", "unit", "file_id"],
    ).set_index(index_start.index.names)

    exp = MovePlan(
        moved_index=exp_moved_index,
        moved_file_map=exp_moved_file_map,
        rewrite_actions=(
            ReWriteAction(
                from_file=file_map_start.loc[2],
                locator=pd.MultiIndex.from_frame(
                    pd.DataFrame(
                        [
                            ("scenario_c", "variable_a", "Mt"),
                        ],
                        columns=["scenario", "variable", "unit"],
                    )
                ),
                to_file=exp_moved_file_map.loc[3],
            ),
        ),
        delete_paths=(file_map_start.loc[1], file_map_start.loc[2]),
    )

    res = make_move_plan(
        index_start=index_start,
        file_map_start=file_map_start,
        data_to_write=data_to_write,
        get_new_data_file_path=db.get_new_data_file_path,
    )

    assert_move_plan_equal(res, exp)



"""
Basic unit tests of `pandas_openscm.indexing`
"""

from __future__ import annotations

import re

import pandas as pd
import pytest

from pandas_openscm.indexing import (
    index_name_aware_lookup,
    multi_index_lookup,
    multi_index_match,
)


def test_unusable_index_levels():
    idx = pd.MultiIndex.from_tuples(
        [
            ("a", "b"),
            ("c", "d"),
        ],
        names=["model", "scenario"],
    )

    locator = pd.MultiIndex.from_tuples(
        [
            ("a", "b", 1),
            ("c", "d", 2),
        ],
        names=["model", "scenario", "run_id"],
    )

    with pytest.raises(
        KeyError,
        match=re.escape(
            "The following levels in `locator` are not in `idx`: ['run_id']. "
            f"{locator.names=} {idx.names=}"
        ),
    ):
        multi_index_match(idx=idx, locator=locator)


def test_index_is_not_multi_index():
    with pytest.raises(
        TypeError,
        match=re.escape(
            "This function is only intended to be used "
            "when `df`'s index is an instance of `MultiIndex`. "
            "Received type(pandas_obj.index)"
            "=<class 'pandas.core.indexes.range.RangeIndex'>"
        ),
    ):
        multi_index_lookup(pd.DataFrame(pd.Index([0, 1])), locator="not used")


def test_index_name_aware_lookup():
    with pytest.raises(
        TypeError,
        match=re.escape(
            "This function is only intended to be used "
            "when `df`'s index is an instance of `MultiIndex`. "
            "Received type(pandas_obj.index)"
            "=<class 'pandas.core.indexes.range.RangeIndex'>"
        ),
    ):
        index_name_aware_lookup(pd.DataFrame(pd.Index([0, 1])), locator="not used")



"""
Functionality for re-writing a database

Mostly used to make way for new data to be written
or to overwrite old data
"""

from __future__ import annotations

import concurrent.futures
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Callable

import numpy as np
import pandas as pd
from attrs import define

from pandas_openscm.db.interfaces import OpenSCMDBDataBackend
from pandas_openscm.index_manipulation import (
    unify_index_levels_check_index_types,
    update_index_from_candidates,
)
from pandas_openscm.indexing import mi_loc, multi_index_match
from pandas_openscm.parallelisation import ParallelOpConfig, apply_op_parallel_progress

if TYPE_CHECKING:
    import pandas.core.indexes.frozen


@define
class MovePlan:
    """Plan for how to move data to make way for an overwrite"""

    moved_index: pd.DataFrame
    """The index once all the data has been moved"""

    moved_file_map: pd.Series[Path]  # type: ignore # pandas confused about ability to support Path
    """The file map once all the data has been moved"""

    rewrite_actions: tuple[ReWriteAction, ...] | None
    """The re-write actions which need to be performed"""

    delete_paths: tuple[Path, ...] | None
    """Paths which can be deleted (after the data has been moved)"""


@define
class ReWriteAction:
    """Description of a re-write action"""

    from_file: Path
    """File from which to load the data"""

    to_file: Path
    """File in which to write the re-written data"""

    locator: pd.MultiIndex
    """Locator which specifies which data to re-write"""


def rewrite_file(
    rewrite_action: ReWriteAction,
    backend: OpenSCMDBDataBackend,
) -> None:
    """
    Re-write a file

    Parameters
    ----------
    rewrite_action
        Re-write action to perform

    backend
        Back-end to use for reading and writing data
    """
    data_all = backend.load_data(rewrite_action.from_file)
    if not backend.preserves_index:
        rewrite_action_names: pandas.core.indexes.frozen.FrozenList = (
            rewrite_action.locator.names  # type: ignore # pandas type hints wrong
        )
        data_all = update_index_from_candidates(
            data_all,
            rewrite_action_names,
        )

    data_rewrite = mi_loc(data_all, rewrite_action.locator)
    backend.save_data(data_rewrite, rewrite_action.to_file)


def rewrite_files(
    rewrite_actions: Iterable[ReWriteAction],
    backend: OpenSCMDBDataBackend,
    parallel_op_config: ParallelOpConfig | None = None,
    progress: bool = False,
    max_workers: int | None = None,
) -> None:
    """
    Re-write a number of files

    Parameters
    ----------
    rewrite_actions
        Re-write actions to perform

    backend
        Backend to use to load and write the files

    parallel_op_config
        Configuration for executing the operation in parallel with progress bars

        If not supplied, we use the values of `progress` and `max_workers`.

    progress
        Should progress bar(s) be used to display the progress of the deletion?

        Only used if `parallel_op_config` is `None`.

    max_workers
        Maximum number of workers to use for parallel processing.

        If supplied, we create an instance of
        [concurrent.futures.ProcessPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor)
        with the provided number of workers.
        A process pool seems to be the sensible default from our experimentation,
        but it is not a universally better choice.
        If you need something else because of how your database is set up,
        simply pass `parallel_op_config`
        rather than using the shortcut of passing `max_workers`.

        If not supplied, the loading is executed serially.

        Only used if `parallel_op_config` is `None`.
    """
    iterable_input: Iterable[ReWriteAction] | list[ReWriteAction] = rewrite_actions

    # Stick the whole thing in a try finally block so we shutdown
    # the parallel pool, even if interrupted, if we created it.
    try:
        if parallel_op_config is None:
            parallel_op_config_use = ParallelOpConfig.from_user_facing(
                progress=progress,
                progress_results_kwargs=dict(desc="File re-writing"),
                progress_parallel_submission_kwargs=dict(
                    desc="Submitting files to the parallel executor"
                ),
                max_workers=max_workers,
                # Process pool by default as basic tests suggest
                # that reading, therefore re-writing, is CPU-bound.
                # See the docs for nuance though.
                parallel_pool_cls=concurrent.futures.ProcessPoolExecutor,
            )
        else:
            parallel_op_config_use = parallel_op_config

        if parallel_op_config_use.progress_results is not None:
            # Wrap in list to force the length to be available to any progress bar.
            # This might be the wrong decision in a weird edge case,
            # but it's convenient enough that I'm willing to take that risk
            iterable_input = list(iterable_input)

        apply_op_parallel_progress(
            func_to_call=rewrite_file,
            iterable_input=iterable_input,
            parallel_op_config=parallel_op_config_use,
            backend=backend,
        )

    finally:
        if parallel_op_config_use.executor_created_in_class_method:
            if parallel_op_config_use.executor is None:  # pragma: no cover
                # Should be impossible to get here
                raise AssertionError

            parallel_op_config_use.executor.shutdown()


def make_move_plan(
    *,
    index_start: pd.DataFrame,
    file_map_start: pd.Series[Path],  # type: ignore # pandas confused about ability to support Path
    data_to_write: pd.DataFrame,
    get_new_data_file_path: Callable[[int], Path],
) -> MovePlan:
    """
    Make a plan for moving data around to make room for new data

    Parameters
    ----------
    index_start
        The starting index

    file_map_start
        The starting file map

    data_to_write
        Data that is going to be written in the database

    Returns
    -------
    :
        Plan for moving data to make room for the new data
    """
    index_start_index_unified, data_to_write_index_unified = (
        unify_index_levels_check_index_types(index_start.index, data_to_write.index)
    )
    in_data_to_write = pd.Series(
        multi_index_match(index_start_index_unified, data_to_write_index_unified),  # type: ignore # pandas type hints confused
        index=index_start.set_index("file_id", append=True).index,
    )

    grouper = in_data_to_write.groupby("file_id")
    no_overwrite = ~grouper.apply(np.any)
    if no_overwrite.all():
        # Don't need to move anything, just return what we started with
        return MovePlan(
            moved_index=index_start,
            moved_file_map=file_map_start,
            rewrite_actions=None,
            delete_paths=None,
        )

    full_overwrite: pd.Series[bool] = grouper.apply(np.all)
    partial_overwrite = ~(full_overwrite | no_overwrite)
    if not partial_overwrite.any():
        # Don't need to move anything,
        # but do no need to delete some files
        # to make way for the parts of the index that will be overwritten
        # (would be even more efficient to just update the file IDs,
        # but that would create a coupling I can't get my head around right now).
        delete_file_ids = full_overwrite.index[full_overwrite]
        delete_paths = file_map_start.loc[delete_file_ids]
        moved_index = index_start[~index_start["file_id"].isin(delete_file_ids)]
        file_map_out = file_map_start.loc[moved_index["file_id"].unique()]

        return MovePlan(
            moved_index=moved_index,
            moved_file_map=file_map_out,
            rewrite_actions=None,
            delete_paths=tuple(delete_paths),
        )

    # Neither nothing to do or only deletions i.e. the fun part.
    to_keep_via_rewrite = partial_overwrite & ~in_data_to_write

    full_overwrite_file_ids = full_overwrite.index[full_overwrite]
    partial_overwrite_file_ids = partial_overwrite.index[partial_overwrite]
    file_ids_to_delete = np.union1d(full_overwrite_file_ids, partial_overwrite_file_ids)
    delete_paths = file_map_start.loc[file_ids_to_delete]

    file_id_map = {}
    max_file_id_start = file_map_start.index.max()
    # Start just with the files that aren't affected by the overwrite
    file_map_out = file_map_start[no_overwrite].copy()
    rewrite_actions_l = []
    for increment, (file_id_old, fiddf) in enumerate(
        # Figure out where to rewrite the data that needs to be rewritten
        to_keep_via_rewrite.loc[to_keep_via_rewrite].groupby("file_id")
    ):
        new_file_id = max_file_id_start + 1 + increment

        file_map_out.loc[new_file_id] = get_new_data_file_path(new_file_id)

        rewrite_actions_l.append(
            ReWriteAction(
                from_file=file_map_start.loc[file_id_old],
                to_file=file_map_out.loc[new_file_id],
                locator=fiddf.index.droplevel("file_id"),
            )
        )
        file_id_map[file_id_old] = new_file_id

    index_keep_via_rewrite = in_data_to_write[
        ~in_data_to_write
        & in_data_to_write.index.get_level_values("file_id").isin(
            partial_overwrite_file_ids
        )
    ].reset_index("file_id")[["file_id"]]

    index_keep_via_rewrite["file_id"] = index_keep_via_rewrite["file_id"].map(
        file_id_map
    )
    if index_keep_via_rewrite["file_id"].isnull().any():  # pragma: no cover
        # Something has gone wrong, everything should be remapped somewhere
        raise AssertionError

    moved_index = pd.concat(
        [
            # Bits of the index which won't be overwritten
            index_start[~index_start["file_id"].isin(file_ids_to_delete)],
            # Bits of the index which are being kept after a rewrite
            index_keep_via_rewrite,
        ]
    )
    res = MovePlan(
        moved_index=moved_index,
        moved_file_map=file_map_out,
        rewrite_actions=tuple(rewrite_actions_l),
        delete_paths=tuple(delete_paths),
    )

    return res



"""
Tests of `pandas_openscm.indexing` and `pd.DataFrame.openscm.mi_loc`
"""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.indexing import (
    index_name_aware_lookup,
    index_name_aware_match,
    multi_index_lookup,
    multi_index_match,
)
from pandas_openscm.testing import create_test_df

try:
    import pandas_indexing as pix
except ImportError:
    pix = None


@pytest.mark.parametrize(
    "start, locator, exp",
    (
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sa", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("mb", "sa", 3),
                ),
                names=["model", "scenario", "id"],
            ),
            [True, False, True, False],
            id="all-levels-covered",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sa", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.MultiIndex.from_tuples(
                (("ma",),),
                names=["model"],
            ),
            [True, True, False, False],
            id="only-first-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sa", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.MultiIndex.from_tuples(
                (("sa",),),
                names=["scenario"],
            ),
            [True, False, True, False],
            id="only-second-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sa", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.MultiIndex.from_tuples(
                (("ma", 1), ("mb", 4)),
                names=["model", "id"],
            ),
            [True, False, False, True],
            id="first-and-third-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sa", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.MultiIndex.from_tuples(
                (("sb", 2), ("sa", 3)),
                names=["scenario", "id"],
            ),
            [False, True, True, False],
            id="second-and-third-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sa", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.MultiIndex.from_tuples(
                (("sb", 2), ("sa", 4)),
                names=["scenario", "id"],
            ),
            [False, True, False, False],
            id="second-and-third-level-not-all-present",
        ),
    ),
)
def test_multi_index_match(start, locator, exp):
    res = multi_index_match(start, locator)
    # # If you want to see what fails with plain pandas, use the below
    # res = start.isin(locator)
    np.testing.assert_equal(res, exp)


def test_multi_index_lookup():
    # Most of the tests are in test_multi_index_match.
    # Hence why there is only one here.
    start = pd.DataFrame(
        np.arange(8).reshape((4, 2)),
        columns=[2010, 2020],
        index=pd.MultiIndex.from_tuples(
            (
                ("ma", "sa", 1),
                ("ma", "sb", 2),
                ("mb", "sa", 3),
                ("mb", "sb", 4),
            ),
            names=["model", "scenario", "id"],
        ),
    )

    locator = pd.MultiIndex.from_tuples(
        (("sb", 2), ("sa", 3), ("sa", 4)),
        names=["scenario", "id"],
    )

    exp = start.iloc[[1, 2], :]

    res = multi_index_lookup(start, locator)

    pd.testing.assert_frame_equal(res, exp)


@pytest.mark.parametrize(
    "start, locator, exp",
    (
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sc", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.Index(["ma", "mb"], name="model"),
            [True, True, True, True],
            id="first-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sc", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.Index(["sa", "sb"], name="scenario"),
            [True, True, False, True],
            id="second-level",
        ),
        pytest.param(
            pd.MultiIndex.from_tuples(
                (
                    ("ma", "sa", 1),
                    ("ma", "sb", 2),
                    ("mb", "sc", 3),
                    ("mb", "sb", 4),
                ),
                names=["model", "scenario", "id"],
            ),
            pd.Index(["sa", "sb"], name="scenario"),
            [True, True, False, True],
            id="third-level",
        ),
    ),
)
def test_index_name_aware_match(start, locator, exp):
    res = index_name_aware_match(start, locator)
    # # # If you want to see what fails with plain pandas, use the below
    # res = start.isin(locator)
    np.testing.assert_equal(res, exp)


def test_index_name_aware_lookup():
    # Most of the tests are in test_index_name_aware_match.
    # Hence why there is only one here.
    start = pd.DataFrame(
        np.arange(8).reshape((4, 2)),
        columns=[2010, 2020],
        index=pd.MultiIndex.from_tuples(
            (
                ("ma", "sa", 1),
                ("ma", "sb", 2),
                ("mb", "sa", 3),
                ("mb", "sb", 4),
            ),
            names=["model", "scenario", "id"],
        ),
    )

    locator = pd.Index((2, 4), name="id")

    exp = start.iloc[[1, 3], :]

    res = index_name_aware_lookup(start, locator)

    pd.testing.assert_frame_equal(res, exp)


@pytest.mark.parametrize(
    "locator",
    (
        pytest.param(["scenario_2", "scenario_1"], id="list"),
        pytest.param(pd.Index(["scenario_2", "scenario_1"]), id="index-no-name"),
        pytest.param(
            ["variable_2", "variable_3"],
            id="list-second-level",
            marks=pytest.mark.xfail(
                reason="pandas looks up the first level rather than variables"
            ),
        ),
        pytest.param(
            pd.Index(["variable_2", "variable_3"]),
            id="index-no-name-second-level",
            marks=pytest.mark.xfail(
                reason="pandas looks up the first level rather than variables"
            ),
        ),
        pytest.param(
            pix.isin(scenario=["scenario_1", "scenario_3"])
            if pix is not None
            else None,
            id="pix_isin",
            marks=pytest.mark.skipif(pix is None, reason="pandas-indexing unavailable"),
        ),
        pytest.param(
            pix.ismatch(
                scenario=[
                    "*1",
                ]
            )
            if pix is not None
            else None,
            id="pix_ismatch",
            marks=pytest.mark.skipif(pix is None, reason="pandas-indexing unavailable"),
        ),
    ),
)
def test_mi_loc_same_as_pandas(locator, setup_pandas_accessor):
    """
    Test pass through in the cases where pass through should happen

    For the cases where there shouldn't be pass through,
    see `test_multi_index_match`
    and `test_index_name_aware_match`.
    """
    start = create_test_df(
        variables=[(f"variable_{i}", "Mt") for i in range(5)],
        n_scenarios=3,
        n_runs=6,
        timepoints=np.arange(1990.0, 2010.0 + 1.0),
    )

    pd.testing.assert_frame_equal(
        start.loc[locator],
        start.openscm.mi_loc(locator),
    )



"""
Tests of `pandas_openscm.parallelisation`
"""

from __future__ import annotations

import concurrent.futures
import multiprocessing

import pytest

from pandas_openscm.parallelisation import ParallelOpConfig, apply_op_parallel_progress
from pandas_openscm.testing import changer

tqdm = pytest.importorskip("tqdm")
tqdm_asyncio = pytest.importorskip("tqdm.asyncio")


def test_from_user_facing_simplest():
    """
    Test if you use `from_user_facing` with the simplest inputs
    """
    res = ParallelOpConfig.from_user_facing(progress=True, max_workers=3)

    results_bar = res.progress_results(["an", "iterable"])
    assert isinstance(results_bar, tqdm_asyncio.tqdm_asyncio)
    assert results_bar.desc == ""

    assert isinstance(res.executor, concurrent.futures.ProcessPoolExecutor)
    assert res.executor._max_workers == 3

    submission_bar = res.progress_parallel_submission(["an", "iterable"])
    assert isinstance(submission_bar, tqdm_asyncio.tqdm_asyncio)
    assert submission_bar.desc == ""

    assert res.executor_created_in_class_method

    res.executor.shutdown()


def test_from_user_facing_progress_bar_kwargs():
    """
    Test if you use `from_user_facing`, controlling some progress bar features
    """
    res = ParallelOpConfig.from_user_facing(
        progress=True,
        progress_results_kwargs=dict(desc="Results"),
        progress_parallel_submission_kwargs=dict(desc="Submission"),
        max_workers=4,
    )

    results_bar = res.progress_results(["an", "iterable"])
    assert isinstance(results_bar, tqdm_asyncio.tqdm_asyncio)
    assert results_bar.desc == "Results"

    assert isinstance(res.executor, concurrent.futures.ProcessPoolExecutor)
    assert res.executor._max_workers == 4

    submission_bar = res.progress_parallel_submission(["an", "iterable"])
    assert isinstance(submission_bar, tqdm_asyncio.tqdm_asyncio)
    assert submission_bar.desc == "Submission"

    assert res.executor_created_in_class_method

    res.executor.shutdown()


def test_from_user_facing_progress_only():
    """
    Test if you use `from_user_facing`, just turning on progress, nothing else
    """
    res = ParallelOpConfig.from_user_facing(progress=True)

    results_bar = res.progress_results(["an", "iterable"])
    assert isinstance(results_bar, tqdm_asyncio.tqdm_asyncio)
    assert results_bar.desc == ""

    assert res.executor is None

    assert res.progress_parallel_submission is None

    assert not res.executor_created_in_class_method


def test_from_user_facing_parallel_only():
    """
    Test if you use `from_user_facing`, just turning on progress, nothing else
    """
    res = ParallelOpConfig.from_user_facing(max_workers=4)

    assert res.progress_results is None

    assert isinstance(res.executor, concurrent.futures.ProcessPoolExecutor)
    assert res.executor._max_workers == 4

    assert res.progress_parallel_submission is None

    assert res.executor_created_in_class_method

    res.executor.shutdown()


apply_op_parallel_progress_args_kwargs = pytest.mark.parametrize(
    "args, kwargs",
    (
        pytest.param(
            [2.0],
            {},
            id="no-kwargs",
        ),
        pytest.param(
            [2.0],
            {"exponent": 1.2},
            id="kwargs",
        ),
    ),
)


@pytest.mark.parametrize(
    "parallel_op_config",
    (
        pytest.param(
            ParallelOpConfig(
                progress_results=None,
                executor=None,
                progress_parallel_submission=None,
            ),
            id="basic",
        ),
        pytest.param(
            ParallelOpConfig.from_user_facing(progress=True),
            id="progress-only",
        ),
        pytest.param(
            ParallelOpConfig(
                progress_results=tqdm.tqdm,
                executor=None,
                progress_parallel_submission=None,
            ),
            id="progress-results-only",
        ),
    ),
)
@apply_op_parallel_progress_args_kwargs
def test_apply_op_parallel_progress_no_executor(parallel_op_config, args, kwargs):
    res = apply_op_parallel_progress(
        changer,
        range(4),
        parallel_op_config,
        *args,
        **kwargs,
    )

    assert isinstance(res, tuple)

    exp = tuple(changer(v, *args, **kwargs) for v in range(4))

    assert res == exp


@pytest.mark.parametrize(
    "executor_cls, executor_cls_kwargs",
    (
        pytest.param(
            concurrent.futures.ProcessPoolExecutor,
            # Have to use spawn within pytest
            dict(mp_context=multiprocessing.get_context("spawn")),
            id="process-pool-spwan",
        ),
        pytest.param(
            concurrent.futures.ThreadPoolExecutor,
            {},
            id="thread-pool",
        ),
    ),
)
@pytest.mark.parametrize(
    "progress_results",
    (
        pytest.param(None, id="no-progress-results"),
        pytest.param(tqdm.tqdm, id="progress-results"),
    ),
)
@pytest.mark.parametrize(
    "progress_parallel_submission",
    (
        pytest.param(None, id="no-progress-parallel-submission"),
        pytest.param(tqdm.tqdm, id="progress-parallel-submission"),
    ),
)
@apply_op_parallel_progress_args_kwargs
def test_apply_op_parallel_progress_our_executor(  # noqa: PLR0913
    executor_cls,
    executor_cls_kwargs,
    progress_results,
    progress_parallel_submission,
    args,
    kwargs,
):
    with executor_cls(**executor_cls_kwargs) as executor:
        parallel_op_config = ParallelOpConfig(
            progress_results=progress_results,
            executor=executor,
            progress_parallel_submission=progress_parallel_submission,
        )

        res = apply_op_parallel_progress(
            changer,
            range(4),
            parallel_op_config,
            *args,
            **kwargs,
        )

    assert isinstance(res, tuple)

    exp = tuple(changer(v, *args, **kwargs) for v in range(4))

    # Order not guaranteed with parallel runs
    assert set(res) == set(exp)



"""
Support for grouping in various ways
"""

from __future__ import annotations

from typing import Any

import pandas as pd


def groupby_except(
    df: pd.DataFrame, non_groupers: str | list[str], observed: bool = True
) -> pd.core.groupby.generic.DataFrameGroupBy[Any]:
    """
    Group by all index levels except specified levels

    This is the inverse of [pd.DataFrame.groupby][pandas.DataFrame.groupby].

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] to group

    non_groupers
        Columns to exclude from the grouping

    observed
        Whether to only return observed combinations or not

    Returns
    -------
    :
        `df`, grouped by all columns except `non_groupers`.
    """
    if isinstance(non_groupers, str):
        non_groupers = [non_groupers]

    return df.groupby(df.index.names.difference(non_groupers), observed=observed)  # type: ignore # pandas-stubs confused


def fix_index_name_after_groupby_quantile(
    df: pd.DataFrame, new_name: str = "quantile", copy: bool = False
) -> pd.DataFrame:
    """
    Fix the index name after performing a `groupby(...).quantile(...)` operation

    By default, pandas doesn't assign a name to the quantile level
    when doing an operation of the form given above.
    This fixes this, but it does assume
    that the quantile level is the only unnamed level in the index.

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] of which to fix the name

    new_name
        New name to give to the quantile column

    copy
        Whether to copy `df` before manipulating the index name

    Returns
    -------
    :
        `df`, with the last level in its index renamed to `new_name`.
    """
    if copy:
        res = df.copy()
    else:
        res = df

    res.index = res.index.rename({None: new_name})

    return res



"""
Parallelisation helpers

Parallelisation is notoriously difficult.
Here we capture some helpers that we have found useful.
If you want to get started with parallelisation,
have a look at some of these resources:

- https://britishgeologicalsurvey.github.io/science/python-forking-vs-spawn/
- https://pythonspeed.com/articles/python-multiprocessing/
- https://joblib.readthedocs.io/en/latest/parallel.html

If you want to understand why I've gone with
[concurrent.futures][] as the default,
have a look at
[this excellent Stack Overflow answer](https://stackoverflow.com/a/60644649)
(tl;dr - [concurrent.futures][] does enough,
the added complexity of the lower-level interfaces isn't worth it).
"""

from __future__ import annotations

import concurrent.futures
from collections.abc import Iterable, Iterator
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar

from attrs import define

from pandas_openscm.exceptions import MissingOptionalDependencyError

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec

    P = ParamSpec("P")
    T = TypeVar("T")
    T_co = TypeVar("T_co", covariant=True)
    U = TypeVar("U")
    V = TypeVar("V")


class ProgressLike(Protocol):
    """A callable that acts like something which creates a progress bar"""

    def __call__(
        self, iterable: Iterable[V], total: int | float | None = None
    ) -> Iterator[V]:
        """
        Create the progress bar

        Parameters
        ----------
        iterable
            Iterable to wrap

        total
            Total number of iterations if known.
        """


def get_tqdm_auto(**kwargs: Any) -> ProgressLike:
    """
    Get a progress bar from [tqdm.auto](https://tqdm.github.io/docs/shortcuts/#tqdmauto).

    Strictly speaking, we use `tqdm.auto.tqdm`.

    Parameters
    ----------
    **kwargs
        Passed to `tqdm.auto.tqdm`.

    Returns
    -------
    :
        Default progress bar.

    Notes
    -----
    This abstraction isn't stricly necessary.
    It's helpful for having this in one place
    and also to only have to do the type hinting fix in one place.
    """
    try:
        import tqdm.auto
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_tqdm_auto", requirement="tqdm"
        ) from exc

    return partial(tqdm.auto.tqdm, **kwargs)  # type: ignore # can't get tqdm and mypy to play nice


@define
class ParallelOpConfig:
    """
    Configuration for a potentially parallel op, potentially with a progress bar(s)
    """

    progress_results: ProgressLike | None = None
    """
    Progress bar to track the results of the op.

    If `None`, no progress bar is displayed for the results of the op.
    """

    # Note: I considered switching the executor for a Protocol here.
    # However, our parallelisation and progress bar display is tightly bound to
    # concurrent.futures' Future class.
    # I figure that, if a user wants to use another pattern,
    # they can and they will probably have other optimisations
    # they want to make too so I'm not going to try and make this too general now.
    executor: concurrent.futures.Executor | None = None
    """
    Executor with which to perform the op.

    If `None`, the op is executed serially.
    """

    progress_parallel_submission: ProgressLike | None = None
    """
    Progress bar to track the submission of the iterable to the parallel executor.

    If `None`, no progress bar is displayed for the results of the op.
    """

    executor_created_in_class_method: bool = False
    """
    Whether `self.executor` was created in a class method

    This can be used to indicate that the user needs to shutdown the executor,
    it was not created from an accessible place.
    """

    @classmethod
    def from_user_facing(
        cls,
        progress: bool = False,
        progress_results_kwargs: dict[str, Any] | None = None,
        progress_parallel_submission_kwargs: dict[str, Any] | None = None,
        max_workers: int | None = None,
        parallel_pool_cls: (
            type[concurrent.futures.ProcessPoolExecutor]
            | type[concurrent.futures.ThreadPoolExecutor]
        ) = concurrent.futures.ProcessPoolExecutor,
    ) -> ParallelOpConfig:
        """
        Initialise from more user-facing arguments

        Parameters
        ----------
        progress
            Should we show progress bars?

        progress_results_kwargs
            Passed to [get_tqdm_auto][(m).] when creating the results bar.

            Only used if `progress`.

        progress_parallel_submission_kwargs
            Passed to [get_tqdm_auto][(m).] when creating the parallel submission bar.

            Only used if `progress` and `max_workers` is not `None`.

        max_workers
            Maximum number of parallel workers.

            If `None`, we set `executor` equal to `None` in the result.

        parallel_pool_cls
            Type of parallel pool executor to use if `max_workers` is not `None`.

        Returns
        -------
        :
            Initialised instance.
        """
        if progress:
            if progress_results_kwargs is None:
                progress_results_kwargs = {}

            progress_results: ProgressLike | None = get_tqdm_auto(
                **progress_results_kwargs
            )

            if max_workers is not None:
                if progress_parallel_submission_kwargs is None:
                    progress_parallel_submission_kwargs = {}

                progress_parallel_submission = get_tqdm_auto(
                    **progress_parallel_submission_kwargs
                )

            else:
                progress_parallel_submission = None

        else:
            progress_results = progress_parallel_submission = None

        if max_workers is not None:
            executor = parallel_pool_cls(max_workers=max_workers)
            executor_created_in_class_method = True

        else:
            executor = None
            executor_created_in_class_method = False

        return cls(
            progress_results=progress_results,
            executor=executor,
            progress_parallel_submission=progress_parallel_submission,
            executor_created_in_class_method=executor_created_in_class_method,
        )


def apply_op_parallel_progress(
    func_to_call: Callable[Concatenate[U, P], T],
    iterable_input: Iterable[U],
    parallel_op_config: ParallelOpConfig,
    *args: P.args,
    **kwargs: P.kwargs,
) -> tuple[T, ...]:
    """
    Apply an operation, potentially in parallel and potentially with progress bars

    Parameters
    ----------
    func_to_call
        Operation to apply to (i.e. the function to call on) `iterable_input`.

    iterable_input
        The input on which to perform `func_to_call`.

        Each element of `iterable_input` is passed to `func_to_call`.

    parallel_op_config
        Configuration with which to execute the potentially parallel process

    *args
        Passed to each call to `func_to_call`.

    **kwargs
        Passed to each call to `func_to_call`.

    Returns
    -------
    :
        Results of each call to `func_to_call`.

    Notes
    -----
    This function does not handle nested progress bars in an obvious way.
    They are possible, but you'll need to set the bar creation stuff up
    appropriately in the layer above calling this function
    (this function doesn't need to know about the location of the bars,
    it just uses whatever bars it is given).

    This function also doesn't handle keeping track of order
    i.e. which outputs map to which inputs.
    If you need this, either add the injection suggested in the code below
    or create another function(
    e.g. `create apply_op_parallel_progress_output_tracker`).
    Our suggestion would be to do the latter first
    and only switch to injection if we find we need more flexibility,
    because the abstractions will likely become hard to follow.
    """
    if parallel_op_config.executor is None:
        # Run serially
        if parallel_op_config.progress_results:
            iterable_input = parallel_op_config.progress_results(iterable_input)

        # If you wanted to make output tracking injectable, something like
        # the below could allow that.
        #
        # res_l = [
        #   output_generator(func_to_call, v, *args, **kwargs) for v in iterable_input
        # ]
        #
        # output_generator would be something like
        # output_generator: Callable[
        #    [Callable[Concatenate[U, P], T], U, P.args, P.kwargs], V
        # ]
        # return type becomes tuple[V, ...]
        #
        # Honestly, I'm not sure it's a helpful abstraction
        # so would just duplicate first
        # and only abstract later if it was really obvious that it was needed.
        res = tuple(func_to_call(v, *args, **kwargs) for v in iterable_input)

        return res

    if parallel_op_config.progress_parallel_submission:
        iterable_input = parallel_op_config.progress_parallel_submission(iterable_input)

    futures = [
        parallel_op_config.executor.submit(
            func_to_call,
            v,
            *args,
            **kwargs,
        )
        for v in iterable_input
    ]

    iterator_results = concurrent.futures.as_completed(futures)
    if parallel_op_config.progress_results:
        iterator_results = parallel_op_config.progress_results(
            iterator_results, total=len(futures)
        )

    res = tuple(ft.result() for ft in iterator_results)

    return res


# # This is how you could abstract out the pattern we repeat elsewhere
# # It's not that helpful though, it just saves line of code.
# # I also can't make the type hints work.
# def hard_to_describe(
#     iterable_input: Iterable[U],
#     func_to_call: Callable[Concatenate[U, P], T],
#     parallel_op_config: ParallelOpConfig | None = None,
#     progress: bool = False,
#     max_workers: int | None = None,
#     default_progress_results_kwargs: dict[str, Any] | None = None,
#     default_progress_parallel_submission_kwargs: dict[str, Any] | None = None,
#     default_parallel_pool_cls: (
#         type[concurrent.futures.ProcessPoolExecutor]
#         | type[concurrent.futures.ThreadPoolExecutor]
#         | None
#     ) = None,
#     *args: P.args,
#     **kwargs: P.kwargs,
# ) -> tuple[T, ...]:
#     iter_or_list: Iterable[U] | list[U] = iterable_input
#
#     if default_progress_results_kwargs is None:
#         default_progress_results_kwargs = {}
#
#     if default_progress_parallel_submission_kwargs is None:
#         default_progress_parallel_submission_kwargs = {}
#
#     # Stick the whole thing in a try finally block so we shutdown
#     # the parallel pool, even if interrupted, if we created it.
#     try:
#         if parallel_op_config is None:
#             if default_parallel_pool_cls is None:
#                 msg = (
#                     f"{parallel_op_config=}, "
#                     "please supply default_parallel_pool_cls"
#                 )
#                 raise TypeError(msg)
#
#             parallel_op_config_use = ParallelOpConfig.from_user_facing(
#                 progress=progress,
#                 progress_results_kwargs=default_progress_results_kwargs,
#                 progress_parallel_submission_kwargs=default_progress_parallel_submission_kwargs,  # noqa: E501
#                 max_workers=max_workers,
#                 parallel_pool_cls=default_parallel_pool_cls,
#             )
#         else:
#             parallel_op_config_use = parallel_op_config
#
#         if parallel_op_config_use.progress_results is not None:
#             # Wrap in list to force the length to be available to any progress bar.
#             # This might be the wrong decision in a weird edge case,
#             # but it's convenient enough that I'm willing to take that risk
#             iter_or_list = list(iter_or_list)
#
#         res = apply_op_parallel_progress(
#             func_to_call=func_to_call,
#             iterable_input=iter_or_list,
#             parallel_op_config=parallel_op_config_use,
#             *args,
#             **kwargs,
#         )
#
#     finally:
#         if parallel_op_config_use.executor_created_in_class_method:
#             if parallel_op_config_use.executor is None:  # pragma: no cover
#                 # Should be impossible to get here
#                 raise AssertionError
#
#             parallel_op_config_use.executor.shutdown()
#
#     return res



"""
Functionality for deleting data
"""

from __future__ import annotations

import concurrent.futures
import os
from collections.abc import Iterable
from pathlib import Path

from pandas_openscm.parallelisation import (
    ParallelOpConfig,
    apply_op_parallel_progress,
)


def delete_files(
    files_to_delete: Iterable[Path],
    parallel_op_config: ParallelOpConfig | None = None,
    progress: bool = False,
    max_workers: int | None = None,
) -> None:
    """
    Delete a number of files

    Parameters
    ----------
    files_to_delete
        Files to delete

    parallel_op_config
        Configuration for executing the operation in parallel with progress bars

        If not supplied, we use the values of `progress` and `max_workers`.

    progress
        Should progress bar(s) be used to display the progress of the deletion?

        Only used if `parallel_op_config` is `None`.

    max_workers
        Maximum number of workers to use for parallel processing.

        If supplied, we create an instance of
        [concurrent.futures.ThreadPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor)
        with the provided number of workers
        (a thread pool makes sense as deletion is I/O-bound).

        If not supplied, the deletions are executed serially.

        Only used if `parallel_op_config` is `None`.
    """
    iterable_input: Iterable[Path] | list[Path] = files_to_delete

    # Stick the whole thing in a try finally block so we shutdown
    # the parallel pool, even if interrupted, if we created it.
    try:
        if parallel_op_config is None:
            parallel_op_config_use = ParallelOpConfig.from_user_facing(
                progress=progress,
                progress_results_kwargs=dict(desc="File deletion"),
                progress_parallel_submission_kwargs=dict(
                    desc="Submitting files to the parallel executor"
                ),
                max_workers=max_workers,
                parallel_pool_cls=concurrent.futures.ThreadPoolExecutor,
            )
        else:
            parallel_op_config_use = parallel_op_config

        if parallel_op_config_use.progress_results is not None:
            # Wrap in list to force the length to be available to any progress bar.
            # This might be the wrong decision in a weird edge case,
            # but it's convenient enough that I'm willing to take that risk
            iterable_input = list(iterable_input)

        apply_op_parallel_progress(
            func_to_call=os.remove,
            iterable_input=iterable_input,
            parallel_op_config=parallel_op_config_use,
        )

    finally:
        if parallel_op_config_use.executor_created_in_class_method:
            if parallel_op_config_use.executor is None:  # pragma: no cover
                # Should be impossible to get here
                raise AssertionError

            parallel_op_config_use.executor.shutdown()



"""
Tests of `pandas_openscm.db.saving`
"""

from __future__ import annotations

import pytest

from pandas_openscm.db.saving import SaveAction, save_file


def test_save_file_unrecognised_save_action_info_kind_error():
    info_kind = 12

    with pytest.raises(NotImplementedError, match=str(info_kind)):
        save_file(
            SaveAction(
                info="not used",
                info_kind=info_kind,
                backend="not used",
                save_path="not used",
            )
        )



"""
In-memory backend

Not very useful in practice, but helpful for testing
"""

from __future__ import annotations

from pathlib import Path
from typing import Literal

import pandas as pd
from attrs import define


@define
class InMemoryDataBackend:
    """
    In-memory data backend
    """

    ext: str = ".in-mem"
    """
    Extension to use with files saved by this backend.
    """

    data: dict[str, pd.DataFrame] | None = None
    """
    Data store
    """

    @property
    def preserves_index(self) -> Literal[True]:
        """
        Whether this backend preserves the index of data upon (de-)serialisation
        """
        return True

    def load_data(self, data_file: Path) -> pd.DataFrame:
        """
        Load a data file

        Parameters
        ----------
        data_file
            File from which to load the data

        Returns
        -------
        :
            Loaded data
        """
        if self.data is None:
            raise TypeError

        return self.data[str(data_file)]

    def save_data(self, data: pd.DataFrame, data_file: Path) -> None:
        """
        Save data to disk

        Parameters
        ----------
        data
            Data to save

        data_file
            File in which to save the data
        """
        if self.data is None:
            self.data = {}

        self.data[str(data_file)] = data
        # Have to do this as, even though it's in-memory,
        # the layer above expects to have files to check, remove etc.
        data_file.touch()


@define
class InMemoryIndexBackend:
    """
    In-memory index backend
    """

    ext: str = ".in-mem"
    """
    Extension to use with files saved by this backend.
    """

    index: pd.DataFrame | None = None
    """Index store"""

    file_map: pd.Series[Path] | None = None  # type: ignore # pandas confused about what it supports
    """File map store"""

    @property
    def preserves_index(self) -> Literal[True]:
        """
        Whether this backend preserves the `pd.MultiIndex` upon (de-)serialisation
        """
        return True

    def load_file_map(self, file_map_file: Path) -> pd.DataFrame:
        """
        Load the file map

        Parameters
        ----------
        file_map_file
            File from which to load the file map

        Returns
        -------
        :
            Loaded file map
        """
        if self.file_map is None:
            raise TypeError

        return self.file_map.to_frame()

    def load_index(self, index_file: Path) -> pd.DataFrame:
        """
        Load the index

        Parameters
        ----------
        index_file
            File from which to load the index

        Returns
        -------
        :
            Loaded index
        """
        if self.index is None:
            raise TypeError

        return self.index

    def save_file_map(
        self,
        file_map: pd.Series[Path],  # type: ignore # pandas confused about what it supports
        file_map_file: Path,
    ) -> None:
        """
        Save the file map to disk

        Parameters
        ----------
        file_map
            File map to save

        file_map_file
            File in which to save the file map
        """
        self.file_map = file_map
        # Have to do this as, even though it's in-memory,
        # the layer above expects to have files to check
        file_map_file.touch()

    def save_index(
        self,
        index: pd.DataFrame,
        index_file: Path,
    ) -> None:
        """
        Save the index to disk

        Parameters
        ----------
        index
            Index to save

        index_file
            File in which to save the index
        """
        self.index = index
        # Have to do this as, even though it's in-memory,
        # the layer above expects to have files to check
        index_file.touch()



"""
Stateful testing of the database with [hypothesis](https://hypothesis.readthedocs.io/en/latest/)

This allows us check that a series of operations on the database
yields the same result, independent of data and index back-ends.
It's not a perfect test, but it is a very helpful one for finding edge cases
and making sure that different combinations of operations all work.
"""

from __future__ import annotations

import random
import shutil
import tempfile
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import (
    EmptyDBError,
    InMemoryDataBackend,
    InMemoryIndexBackend,
    OpenSCMDB,
)
from pandas_openscm.testing import (
    assert_frame_alike,
    get_db_data_backends,
    get_db_index_backends,
)

pytest.importorskip("filelock")
hypothesis = pytest.importorskip("hypothesis")
hypothesis_stateful = pytest.importorskip("hypothesis.stateful")
pytest.importorskip("pandas_indexing")
pytestmark = pytest.mark.slow


def get_new_data(
    *,
    n_ts_options: tuple[int, ...],
    timepoint_options: tuple[np.typing.NDArray[float], ...],
    metadata_options: tuple[tuple[str, ...]],
    rng: np.random.Generator,
    data_existing: pd.DataFrame | None,
) -> pd.DataFrame:
    n_ts = random.choice(n_ts_options)  # noqa: S311
    timepoints = random.choice(timepoint_options)  # noqa: S311
    metadata_cols = random.choice(metadata_options)  # noqa: S311

    data_vals = rng.random((n_ts, timepoints.size))
    multi_index_full = []
    n_draws = int(np.ceil(n_ts ** (1 / len(metadata_cols))))
    for col in metadata_cols:
        if data_existing is None:
            min_index = 0
        elif col in data_existing.index.names:
            min_index = (
                max(
                    int(v.replace(f"{col}_", "")) if isinstance(v, str) else 0
                    for v in data_existing.pix.unique(col)
                )
                + 1
            )
        else:
            min_index = 0

        col_vals = [f"{col}_{i}" for i in range(min_index, min_index + n_draws)]
        multi_index_full.append(col_vals)

    data_index = pd.MultiIndex.from_product(multi_index_full, names=metadata_cols)
    # Get the number of samples we're interested in
    data_index = data_index[random.sample(range(data_index.shape[0]), n_ts)]

    new_data = pd.DataFrame(data_vals, index=data_index, columns=timepoints)

    return new_data


class DBMofidierBase(hypothesis.stateful.RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        self.data_exp = None
        self.rng = np.random.default_rng()
        for db in self.dbs:
            db.db_dir.mkdir(exist_ok=True, parents=True)

    def teardown(self):
        for db in self.dbs:
            shutil.rmtree(db.db_dir)

    @hypothesis.stateful.rule()
    def delete(self):
        self.data_exp = None
        for db in self.dbs:
            # Should work irrespective of whether
            # there is anything to delete or not
            db.delete()

    @hypothesis.stateful.rule()
    def add_new_data(self):
        new_data = get_new_data(
            n_ts_options=self.n_ts_options,
            timepoint_options=self.timepoint_options,
            metadata_options=self.metadata_options,
            rng=self.rng,
            data_existing=self.data_exp,
        )

        for db in self.dbs:
            # Should be no overlap hence no overwrite needed
            db.save(new_data)

        if self.data_exp is None:
            self.data_exp = new_data
        else:
            self.data_exp = pd.concat(
                v.dropna(axis="rows", how="all")
                for v in self.data_exp.align(new_data, axis="rows")
            )

    @hypothesis.stateful.rule()
    def add_new_data_grouped(self):
        metadata_options = self.metadata_options

        new_data = get_new_data(
            n_ts_options=self.n_ts_options,
            timepoint_options=self.timepoint_options,
            metadata_options=metadata_options,
            rng=self.rng,
            data_existing=self.data_exp,
        )

        groupby = random.sample(new_data.index.names, len(new_data.index.names) - 1)
        for db in self.dbs:
            # Should be no overlap hence no overwrite needed
            db.save(new_data, groupby=groupby)

        if self.data_exp is None:
            self.data_exp = new_data
        else:
            self.data_exp = pd.concat(
                v.dropna(axis="rows", how="all")
                for v in self.data_exp.align(new_data, axis="rows")
            )

    @hypothesis.stateful.precondition(lambda self: self.data_exp is not None)
    @hypothesis.stateful.rule()
    def add_fully_overlapping_data(self):
        existing_idx = self.dbs[0].load_index()
        to_dup_idx = existing_idx[
            existing_idx["file_id"] == existing_idx["file_id"].min()
        ].index

        to_dup_loc = self.data_exp.index.isin(to_dup_idx)
        dup = self.data_exp.loc[to_dup_loc, :] * 1.1

        for db in self.dbs:
            # Should not need warn_on_partial_overwrite
            # here as we're doing a complete overwrite.
            db.save(dup, allow_overwrite=True)

        self.data_exp = pd.concat(
            [
                self.data_exp.loc[~to_dup_loc, :],
                dup,
            ]
        )

    @hypothesis.stateful.precondition(lambda self: self.data_exp is not None)
    @hypothesis.stateful.rule()
    def add_partially_overlapping_data(self):
        existing_idx = self.dbs[0].load_index()

        file_ids_unique = existing_idx["file_id"].unique()
        n_file_ids = len(file_ids_unique)
        to_dup_idx = None
        for i in range(min(3, n_file_ids)):
            tmp = existing_idx[existing_idx["file_id"] == file_ids_unique[i]].index
            tmp = tmp[: min(3, tmp.size)]
            if to_dup_idx is None:
                to_dup_idx = tmp
            else:
                to_dup_idx = to_dup_idx.append(tmp)

        to_dup_loc = self.data_exp.index.isin(to_dup_idx)
        dup = self.data_exp.loc[to_dup_loc, :] * 1.1

        for db in self.dbs:
            db.save(dup, allow_overwrite=True, warn_on_partial_overwrite=False)

        self.data_exp = pd.concat(
            [
                self.data_exp.loc[~to_dup_loc, :],
                dup,
            ]
        )

    @hypothesis.stateful.invariant()
    def all_db_index_are_multiindex(self):
        if self.data_exp is None:
            return

        assert isinstance(self.data_exp.index, pd.MultiIndex)
        for db in self.dbs:
            try:
                index = db.load_index()
                assert isinstance(index.index, pd.MultiIndex)
            except AssertionError as exc:
                msg = (
                    f"{type(db.backend_data).__name__=}"
                    f"{type(db.backend_index).__name__=}"
                )
                raise AssertionError(msg) from exc

    @hypothesis.stateful.invariant()
    def all_dbs_consistent_with_expected(self):
        for db in self.dbs:
            try:
                if self.data_exp is not None:
                    loaded = db.load(out_columns_type=self.data_exp.columns.dtype)
                    assert isinstance(loaded.index, pd.MultiIndex)
                    assert_frame_alike(loaded, self.data_exp)

                    loaded_metadata = db.load_metadata()
                    assert isinstance(loaded_metadata, pd.MultiIndex)
                    loaded_comparison = (
                        loaded_metadata.to_frame(index=False)
                        .fillna("i_was_nan")
                        .replace("nan", "i_was_nan")
                        .sort_values(self.data_exp.index.names)
                        .reset_index(drop=True)
                    )
                    exp_comparison = (
                        self.data_exp.index.to_frame(index=False)
                        .fillna("i_was_nan")
                        .sort_values(self.data_exp.index.names)
                        .reset_index(drop=True)
                    )
                    pd.testing.assert_frame_equal(
                        loaded_comparison, exp_comparison, check_like=True
                    )

                else:
                    with pytest.raises(EmptyDBError):
                        db.load()

            except AssertionError as exc:
                msg = (
                    f"{type(db.backend_data).__name__=}"
                    f"{type(db.backend_index).__name__=}"
                )
                raise AssertionError(msg) from exc


@hypothesis.settings(max_examples=5)
class DBMofidier(DBMofidierBase):
    dbs = tuple(
        OpenSCMDB(
            backend_data=backend_data(),
            backend_index=backend_index(),
            db_dir=Path(tempfile.mkdtemp()),
        )
        for backend_data in get_db_data_backends()
        for backend_index in get_db_index_backends()
    )
    n_ts_options = (1, 3)
    timepoint_options = (
        np.arange(1995.0, 2005.0 + 1.0),
        np.arange(2000.0, 2020.0 + 1.0),
        np.arange(2015.0, 2025.0 + 1.0),
    )
    metadata_options = (
        ("variable", "unit"),
        ("run_id", "variable", "unit"),
    )


@hypothesis.settings(max_examples=20)
class DBMofidierInMemory(DBMofidierBase):
    dbs = tuple(
        OpenSCMDB(
            backend_data=backend_data(),
            backend_index=backend_index(),
            db_dir=Path(tempfile.mkdtemp()),
        )
        for backend_data in [InMemoryDataBackend]
        for backend_index in [InMemoryIndexBackend]
    )
    n_ts_options = (1, 3, 5, 10)
    timepoint_options = (
        np.arange(2000.0, 2020.0 + 1.0),
        np.arange(2000.0, 2010.0 + 1.0, 2.0),
        np.arange(2010.0, 2020.0 + 1.0, 5.0),
        np.arange(1995.0, 2005.0 + 1.0),
        np.arange(2015.0, 2025.0 + 1.0),
    )
    metadata_options = (
        ("variable", "unit"),
        ("run_id", "variable", "unit"),
        ("scenario", "variable", "unit", "run_id"),
    )


DBModifierTest = pytest.mark.superslow(DBMofidier.TestCase)
DBModifierInMemoryTest = DBMofidierInMemory.TestCase



"""
Tests of `pandas_openscm.db.rewriting`
"""

from __future__ import annotations

from pathlib import Path

import pandas as pd
import pytest

from pandas_openscm.db import InMemoryDataBackend, InMemoryIndexBackend, OpenSCMDB
from pandas_openscm.db.rewriting import make_move_plan


def test_make_move_plan_index_start_not_multi_error(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=None,  # not used
    )

    with pytest.raises(TypeError):
        make_move_plan(
            index_start=pd.DataFrame(["a", "b", "c"], pd.Index([1, 2, 3])),
            file_map_start="not used",
            data_to_write="not used",
            get_new_data_file_path=db.get_new_data_file_path,
        )



"""
Definition of our key [OpenSCMDB][(m).] class
"""

from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pandas as pd
from attrs import define, field

from pandas_openscm.db.deleting import delete_files
from pandas_openscm.db.interfaces import OpenSCMDBDataBackend, OpenSCMDBIndexBackend
from pandas_openscm.db.loading import (
    load_data,
    load_db_file_map,
    load_db_index,
    load_db_metadata,
)
from pandas_openscm.db.reader import OpenSCMDBReader
from pandas_openscm.db.rewriting import make_move_plan, rewrite_files
from pandas_openscm.db.saving import save_data
from pandas_openscm.exceptions import MissingOptionalDependencyError
from pandas_openscm.index_manipulation import unify_index_levels_check_index_types
from pandas_openscm.indexing import multi_index_match
from pandas_openscm.parallelisation import (
    ParallelOpConfig,
    ProgressLike,
)

if TYPE_CHECKING:
    import filelock
    import pandas_indexing as pix


class AlreadyInDBError(ValueError):
    """
    Raised when saving data would overwrite data which is already in the database
    """

    def __init__(self, already_in_db: pd.DataFrame) -> None:
        """
        Initialise the error

        Parameters
        ----------
        already_in_db
            data that is already in the database
        """
        error_msg = (
            "The following rows are already in the database:\n"
            f"{already_in_db.index.to_frame(index=False)}"
        )
        super().__init__(error_msg)


class EmptyDBError(ValueError):
    """
    Raised when trying to access data from a database that is empty
    """

    def __init__(self, db: OpenSCMDB) -> None:
        """
        Initialise the error

        Parameters
        ----------
        db
            The database
        """
        error_msg = f"The database is empty: {db=}"
        super().__init__(error_msg)


@define
class OpenSCMDB:
    """
    Database for storing OpenSCM-style data

    This class is focussed on backends that use files as their storage.
    If you had a different database backend,
    you might make different choices.
    We haven't thought through those use cases
    hence aren't sure how much effort
    would be required to make something truly backend agnostic.
    """

    backend_data: OpenSCMDBDataBackend = field(kw_only=True)
    """
    The backend for (de-)serialising data (from) to disk
    """

    backend_index: OpenSCMDBIndexBackend = field(kw_only=True)
    """
    The backend for (de-)serialising the database index (from) to disk
    """

    db_dir: Path = field(kw_only=True)
    """
    Path in which the database is stored

    Both the index and the data files will be written in this directory.
    """

    index_file_lock: filelock.BaseFileLock = field(kw_only=True)
    """
    Lock for the index file
    """
    # Note to devs: filelock releases the lock when __del__ is called
    # (i.e. when the lock instance is garbage collected).
    # Hence, you have to keep a reference to this around
    # if you want it to do anything.
    # For a while, we made this a property that created the lock when requested.
    # That was super confusing as, if the reference to the created lock wasn't kept,
    # the lock would immediately be released.

    @index_file_lock.default
    def default_index_file_lock(self) -> filelock.BaseFileLock:
        """Get default lock for the back-end's index file"""
        try:
            import filelock
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "default_index_file_lock", requirement="filelock"
            ) from exc

        return filelock.FileLock(self.index_file_lock_path)

    @property
    def file_map_file(self) -> Path:
        """
        The file in which the file map is stored

        The file map stores the mapping from file_id
        to file path.

        Returns
        -------
        :
            Path to the file map file
        """
        return self.db_dir / f"filemap{self.backend_index.ext}"

    @property
    def index_file(self) -> Path:
        """
        The file in which the database's index is stored

        Returns
        -------
        :
            Path to the index file
        """
        return self.db_dir / f"index{self.backend_index.ext}"

    @property
    def index_file_lock_path(self) -> Path:
        """Path to the lock file for the back-end's index file"""
        return self.index_file.parent / f"{self.index_file.name}.lock"

    @property
    def is_empty(self) -> bool:
        """
        Whether the database is empty or not

        Returns
        -------
        :
            `True` if the database is empty, `False` otherwise
        """
        return not self.index_file.exists()

    def create_reader(
        self,
        *,
        lock: bool | filelock.BaseFileLock | None = True,
        index_file_lock: filelock.BaseFileLock | None = None,
    ) -> OpenSCMDBReader:
        """
        Create a database reader

        Parameters
        ----------
        lock
            Lock to give to the reader.

            If `True`, we create a new lock for the database, such that,
            if the reader is holding the lock,
            no operations can be performed on the database.

            If `False`, the reader is not given any lock.

        index_file_lock
            Lock for the database's index file

            Used while loading the index from disk.

            If not supplied, we use [self.index_file_lock][(c)].

        Returns
        -------
        :
            Database reader
        """
        if isinstance(lock, bool):
            if lock:
                try:
                    import filelock
                except ImportError as exc:
                    raise MissingOptionalDependencyError(  # noqa: TRY003
                        "create_reader(..., lock=True, ...)", requirement="filelock"
                    ) from exc

                # Create a new lock for the reader
                lock = filelock.FileLock(self.index_file_lock_path)

            else:
                # Convert to None
                lock = None

        db_index = self.load_index(index_file_lock=index_file_lock)
        db_file_map = self.load_file_map(index_file_lock=index_file_lock)

        res = OpenSCMDBReader(
            backend_data=self.backend_data,
            db_index=db_index,
            db_file_map=db_file_map,
            lock=lock,
        )

        return res

    def delete(
        self,
        *,
        index_file_lock: filelock.BaseFileLock | None = None,
        parallel_op_config: ParallelOpConfig | None = None,
        progress: bool = False,
        max_workers: int | None = None,
    ) -> None:
        """
        Delete all data in the database

        Parameters
        ----------
        index_file_lock
            Lock for the database's index file

            If not supplied, we use [self.index_file_lock][(c)].

        parallel_op_config
            Configuration for executing the operation in parallel with progress bars

            If not supplied, we use the values of `progress` and `max_workers`.

        progress
            Should progress bar(s) be used to display the progress of the deletion?

            Only used if `parallel_op_config` is `None`.

        max_workers
            Maximum number of workers to use for parallel processing.

            If supplied, we create an instance of
            [concurrent.futures.ThreadPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor)
            with the provided number of workers
            (a thread pool makes sense as deletion is I/O-bound).

            If not supplied, the deletions are executed serially.

            Only used if `parallel_op_config` is `None`.
        """
        if index_file_lock is None:
            index_file_lock = self.index_file_lock

        with index_file_lock:
            files_to_delete = {
                *self.db_dir.glob(f"*{self.backend_data.ext}"),
                *self.db_dir.glob(f"*{self.backend_index.ext}"),
            }
            delete_files(
                files_to_delete=files_to_delete,
                parallel_op_config=parallel_op_config,
                progress=progress,
                max_workers=max_workers,
            )

    def get_new_data_file_path(self, file_id: int) -> Path:
        """
        Get the path in which to write a new data file

        Parameters
        ----------
        file_id
            ID to associate with the file

        Returns
        -------
        :
            File in which to write the new data

        Raises
        ------
        FileExistsError
            A file already exists for the given `file_id`
        """
        file_path = self.db_dir / f"{file_id}{self.backend_data.ext}"

        if file_path.exists():
            raise FileExistsError(file_path)

        return file_path

    def load(  # noqa: PLR0913
        self,
        selector: pd.Index[Any] | pd.MultiIndex | pix.selectors.Selector | None = None,
        *,
        index_file_lock: filelock.BaseFileLock | None = None,
        out_columns_type: type | None = None,
        parallel_op_config: ParallelOpConfig | None = None,
        progress: bool = False,
        max_workers: int | None = None,
    ) -> pd.DataFrame:
        """
        Load data

        Parameters
        ----------
        selector
            Selector to use to choose the data to load

        index_file_lock
            Lock for the database's index file

            If not supplied, we use [self.index_file_lock][(c)].

        out_columns_type
            Type to set the output columns to.

            If not supplied, we don't set the output columns' type.

        parallel_op_config
            Configuration for executing the operation in parallel with progress bars

            If not supplied, we use the values of `progress` and `max_workers`.

        progress
            Should progress bar(s) be used to display the progress of the deletion?

            Only used if `parallel_op_config` is `None`.

        max_workers
            Maximum number of workers to use for parallel processing.

            If supplied, we create an instance of
            [concurrent.futures.ProcessPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor)
            with the provided number of workers.
            A process pool seems to be the sensible default from our experimentation,
            but it is not a universally better choice.
            If you need something else because of how your database is set up,
            simply pass `parallel_op_config`
            rather than using the shortcut of passing `max_workers`.

            If not supplied, the loading is executed serially.

            Only used if `parallel_op_config` is `None`.

        Returns
        -------
        :
            Loaded data

        Raises
        ------
        EmptyDBError
            The database is empty
        """
        if self.is_empty:
            raise EmptyDBError(self)

        if index_file_lock is None:
            index_file_lock = self.index_file_lock

        with index_file_lock:
            file_map = self.load_file_map(index_file_lock=index_file_lock)
            index = self.load_index(index_file_lock=index_file_lock)

            res = load_data(
                backend_data=self.backend_data,
                db_index=index,
                db_file_map=file_map,
                selector=selector,
                out_columns_type=out_columns_type,
                parallel_op_config=parallel_op_config,
                progress=progress,
                max_workers=max_workers,
            )

        return res

    def load_file_map(
        self,
        *,
        index_file_lock: filelock.BaseFileLock | None = None,
    ) -> pd.Series[Path]:  # type: ignore # pandas type hints confused about what they support
        """
        Load the file map

        Parameters
        ----------
        index_file_lock
            Lock for the database's index file

            If not supplied, we use [self.index_file_lock][(c)].

        Returns
        -------
        :
            Map from file ID to file path

        Raises
        ------
        EmptyDBError
            The database is empty
        """
        if self.is_empty:
            raise EmptyDBError(self)

        if index_file_lock is None:
            index_file_lock = self.index_file_lock

        with index_file_lock:
            file_map = load_db_file_map(
                backend_index=self.backend_index, file_map_file=self.file_map_file
            )

        return file_map

    def load_index(
        self,
        *,
        index_file_lock: filelock.BaseFileLock | None = None,
    ) -> pd.DataFrame:
        """
        Load the index

        Parameters
        ----------
        index_file_lock
            Lock for the database's index file

            If not supplied, we use [self.index_file_lock][(c)].

        Returns
        -------
        :
            Database index

        Raises
        ------
        EmptyDBError
            The database is empty
        """
        if self.is_empty:
            raise EmptyDBError(self)

        if index_file_lock is None:
            index_file_lock = self.index_file_lock

        with index_file_lock:
            index = load_db_index(
                backend_index=self.backend_index,
                index_file=self.index_file,
            )

        return index

    def load_metadata(
        self,
        *,
        index_file_lock: filelock.BaseFileLock | None = None,
    ) -> pd.MultiIndex:
        """
        Load the database's metadata

        Parameters
        ----------
        index_file_lock
            Lock for the database's index file

            If not supplied, we use [self.index_file_lock][(c)].

        Returns
        -------
        :
            Loaded metadata
        """
        if not self.index_file.exists():
            raise EmptyDBError(self)

        if index_file_lock is None:
            index_file_lock = self.index_file_lock

        with index_file_lock:
            metadata = load_db_metadata(
                backend_index=self.backend_index, index_file=self.index_file
            )

        return metadata

    def save(  # noqa: PLR0913
        self,
        data: pd.DataFrame,
        *,
        index_file_lock: filelock.BaseFileLock | None = None,
        groupby: list[str] | None = None,
        allow_overwrite: bool = False,
        warn_on_partial_overwrite: bool = True,
        progress_grouping: ProgressLike | None = None,
        parallel_op_config_save: ParallelOpConfig | None = None,
        parallel_op_config_delete: ParallelOpConfig | None = None,
        parallel_op_config_rewrite: ParallelOpConfig | None = None,
        progress: bool = False,
        max_workers: int | None = None,
    ) -> None:
        """
        Save data into the database

        Parameters
        ----------
        data
            Data to add to the database

        index_file_lock
            Lock for the database's index file

            If not supplied, we use [self.index_file_lock][(c)].

        groupby
            Metadata columns to use to group the data.

            If not supplied, we save all the data in a single file.

        allow_overwrite
            Should overwrites of data that is already in the database be allowed?

            If this is `True`, there is a risk that, if interrupted halfway through,
            you can end up with duplicate data in your database
            or some other odd broken state.

        warn_on_partial_overwrite
            Should a warning be raised if a partial overwrite will occur?

            This is on by default so that users
            are warned about the slow operation of re-writing.

        progress_grouping
            Progress bar to use when grouping the data

            If not supplied, we use the values of `progress` and `max_workers`.

        parallel_op_config_save
            Parallel op configuration for executing save operations

            If not supplied, we use the values of `progress` and `max_workers`.

        parallel_op_config_delete
            Parallel op configuration for executing any needed delete operations

            If not supplied, we use the values of `progress` and `max_workers`.

        parallel_op_config_rewrite
            Parallel op configuration for executing any needed re-write operations

            If not supplied, we use the values of `progress` and `max_workers`.

        progress
            Should progress bar(s) be used to display the progress of the various steps?

            Only used if the corresponding `parallel_op_config_*` variable
            for the operation is `None`.

        max_workers
            Maximum number of workers to use for parallel processing.

            If supplied, we create instances of
            [concurrent.futures.Executor][]
            with the provided number of workers
            (the exact kind of executor depends on the operation).

            If not supplied, the operations are executed serially.

            Only used if the corresponding `parallel_op_config_*` variable
            for the operation is `None`.
        """
        if not isinstance(data.index, pd.MultiIndex):
            msg = (
                "`data.index` must be an instance of `pd.MultiIndex`. "
                f"Received {type(data.index)=}"
            )
            raise TypeError(msg)

        if data.index.duplicated().any():
            duplicate_rows = data.index.duplicated(keep=False)
            duplicates = data.loc[duplicate_rows, :]
            msg = (
                "`data` contains rows with the same metadata. "
                f"duplicates=\n{duplicates}"
            )

            raise ValueError(msg)

        if index_file_lock is None:
            index_file_lock = self.index_file_lock

        with index_file_lock:
            if self.is_empty:
                move_plan = None
                index_non_data = None
                file_map_non_data = None
                min_file_id = 0

            else:
                file_map_db = self.load_file_map(index_file_lock=index_file_lock)
                index_db = self.load_index(index_file_lock=index_file_lock)
                if not allow_overwrite:
                    data_index_unified, index_db_index_unified = (
                        unify_index_levels_check_index_types(data.index, index_db.index)
                    )
                    overwrite_required = multi_index_match(
                        data_index_unified, index_db_index_unified
                    )

                    if overwrite_required.any():
                        data_to_write_already_in_db = data.loc[overwrite_required, :]
                        raise AlreadyInDBError(
                            already_in_db=data_to_write_already_in_db
                        )

                move_plan = make_move_plan(
                    index_start=index_db,
                    file_map_start=file_map_db,
                    data_to_write=data,
                    get_new_data_file_path=self.get_new_data_file_path,
                )

                # As needed, re-write files without deleting the old files
                if move_plan.rewrite_actions is not None:
                    if warn_on_partial_overwrite:
                        msg = (
                            "Overwriting the data will require re-writing. "
                            "This may be slow. "
                            "If that is an issue, the way to solve it "
                            "is to update your workflow to ensure "
                            "that you are not overwriting data "
                            "or are only overwriting entire files."
                        )
                        warnings.warn(msg)

                    rewrite_files(
                        move_plan.rewrite_actions,
                        backend=self.backend_data,
                        parallel_op_config=parallel_op_config_rewrite,
                        progress=progress,
                        max_workers=max_workers,
                    )

                # Write the new data
                current_largest_file_id = file_map_db.index.max()
                if not move_plan.moved_file_map.empty:
                    current_largest_file_id = max(
                        move_plan.moved_file_map.index.max(), current_largest_file_id
                    )

                index_non_data = move_plan.moved_index
                file_map_non_data = move_plan.moved_file_map
                min_file_id = current_largest_file_id + 1

            save_data(
                data,
                backend_data=self.backend_data,
                get_new_data_file_path=self.get_new_data_file_path,
                backend_index=self.backend_index,
                index_file=self.index_file,
                file_map_file=self.file_map_file,
                index_non_data=index_non_data,
                file_map_non_data=file_map_non_data,
                min_file_id=min_file_id,
                groupby=groupby,
                progress_grouping=progress_grouping,
                parallel_op_config=parallel_op_config_save,
                progress=progress,
                max_workers=max_workers,
            )

            # As needed, delete files.
            # We delete files last to minimise the risk of losing data
            # (might end up with double if we get interrupted here,
            # but that is better than zero).
            if move_plan is not None and move_plan.delete_paths is not None:
                delete_files(
                    files_to_delete=move_plan.delete_paths,
                    parallel_op_config=parallel_op_config_delete,
                    progress=progress,
                    max_workers=max_workers,
                )



"""
Plotting
"""

from __future__ import annotations

import warnings
from collections.abc import Collection, Iterable, Iterator, Mapping
from functools import partial
from itertools import cycle
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Literal,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import pandas as pd
from attrs import define, field

from pandas_openscm.exceptions import MissingOptionalDependencyError
from pandas_openscm.grouping import (
    fix_index_name_after_groupby_quantile,
    groupby_except,
)

if TYPE_CHECKING:
    import attr
    import matplotlib
    import pint
    from typing_extensions import TypeAlias

    COLOUR_VALUE_LIKE: TypeAlias = Union[
        Union[
            str,
            tuple[float, float, float],
            tuple[float, float, float, float],
            tuple[Union[tuple[float, float, float], str], float],
            tuple[tuple[float, float, float, float], float],
        ],
    ]
    """Type that allows a colour to be specified in matplotlib"""

    T = TypeVar("T")

    class PALETTE_LIKE(
        Generic[T],
        Mapping[T, COLOUR_VALUE_LIKE],
    ):
        """Palette-like type"""

    # PALETTE_LIKE: TypeAlias = dict[
    #     Any, Union[COLOUR_VALUE_LIKE, tuple[COLOUR_VALUE_LIKE, float]]
    # ]

    DASH_VALUE_LIKE: TypeAlias = Union[str, tuple[float, tuple[float, ...]]]
    """Types that allow a dash to be specified in matplotlib"""

    QUANTILES_PLUMES_LIKE: TypeAlias = tuple[
        Union[tuple[float, float], tuple[tuple[float, float], float]], ...
    ]
    """Type that quantiles and the alpha to use for plotting their line/plume"""

    NP_ARRAY_OF_FLOAT_OR_INT = np.typing.NDArray[
        Union[np.floating[Any], np.integer[Any]]
    ]
    """Numpy array like"""

    PINT_NUMPY_ARRAY: TypeAlias = pint.facets.numpy.quantity.NumpyQuantity[
        NP_ARRAY_OF_FLOAT_OR_INT
    ]
    """
    Type alias for a pint quantity that wraps a numpy array

    No shape hints because that doesn't seem to be supported by numpy yet.
    """


class MissingQuantileError(KeyError):
    """
    Raised when a quantile(s) is missing from a [pd.DataFrame][pandas.DataFrame]
    """

    def __init__(
        self,
        available_quantiles: Collection[float],
        missing_quantiles: Collection[float],
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        available_quantiles
            Available quantiles

        missing_quantiles
            Missing quantiles
        """
        error_msg = (
            f"The folllowing quantiles are missing: {missing_quantiles=}. "
            f"{available_quantiles=}"
        )
        super().__init__(error_msg)


def get_quantiles(
    quantiles_plumes: QUANTILES_PLUMES_LIKE,
) -> np.typing.NDArray[np.floating[Any]]:
    """
    Get just the quantiles from a [QUANTILES_PLUMES_LIKE][(m).]

    Parameters
    ----------
    quantiles_plumes
        Quantiles-plumes definition

    Returns
    -------
    :
        Quantiles to be used in plotting
    """
    quantiles_l = []
    for quantile_plot_def in quantiles_plumes:
        q_def = quantile_plot_def[0]
        if isinstance(q_def, float):
            quantiles_l.append(q_def)
        else:
            for q in q_def:
                quantiles_l.append(q)

    return np.unique(np.array(quantiles_l))  # type: ignore # numpy and mypy not playing nice


def get_pdf_from_pre_calculated(
    in_df: pd.DataFrame,
    *,
    quantiles: Iterable[float],
    quantile_col: str,
) -> pd.DataFrame:
    """
    Get a [pd.DataFrame][pandas.DataFrame] for plotting from pre-calculated quantiles

    Parameters
    ----------
    in_df
        Input [pd.DataFrame][pandas.DataFrame]

    quantiles
        Quantiles to grab

    quantile_col
        Name of the index column in which quantile information is stored

    Returns
    -------
    :
        [pd.DataFrame][pandas.DataFrame] to use for plotting.

    Raises
    ------
    MissingQuantileError
        One of the quantiles in `quantiles` is not available in `in_df`.
    """
    missing_quantiles = []
    available_quantiles = in_df.index.get_level_values(quantile_col).unique().tolist()
    for qt in quantiles:
        if qt not in available_quantiles:
            missing_quantiles.append(qt)

    if missing_quantiles:
        raise MissingQuantileError(available_quantiles, missing_quantiles)

    # otherwise, have what we need
    pdf = in_df.loc[in_df.index.get_level_values(quantile_col).isin(quantiles)]

    return pdf


def extract_single_unit(df: pd.DataFrame, unit_var: str) -> str:
    """
    Extract the unit of the data, expecting there to only be one unit

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] from which to get the unit

    unit_var
        Variable/column in the multi-index which holds unit information

    Returns
    -------
    :
        Unit of the data

    Raises
    ------
    AssertionError
        The data has more than one unit
    """
    units = df.index.get_level_values(unit_var).unique().tolist()
    if len(units) != 1:
        raise AssertionError(units)

    return cast(str, units[0])


@overload
def get_values_line(
    pdf: pd.DataFrame,
    *,
    unit_aware: Literal[False],
    unit_var: str | None,
    time_units: str | None,
) -> tuple[NP_ARRAY_OF_FLOAT_OR_INT, NP_ARRAY_OF_FLOAT_OR_INT]: ...


@overload
def get_values_line(
    pdf: pd.DataFrame,
    *,
    unit_aware: Literal[True] | pint.facets.PlainRegistry,
    unit_var: str | None,
    time_units: str | None,
) -> tuple[PINT_NUMPY_ARRAY, PINT_NUMPY_ARRAY]: ...


def get_values_line(
    pdf: pd.DataFrame,
    *,
    unit_aware: bool | pint.facets.PlainRegistry,
    unit_var: str | None,
    time_units: str | None,
) -> (
    tuple[NP_ARRAY_OF_FLOAT_OR_INT, NP_ARRAY_OF_FLOAT_OR_INT]
    | tuple[PINT_NUMPY_ARRAY, PINT_NUMPY_ARRAY]
):
    """
    Get values for plotting a line

    Parameters
    ----------
    pdf
        [pd.DataFrame][pandas.DataFrame] from which to get the values

    unit_aware
        Should the values be unit-aware?

        If `True`, we use the default application registry
        (retrieved with [pint.get_application_registry][]).
        Otherwise, a [pint.facets.PlainRegistry][] can be supplied and will be used.

    unit_var
        Variable/column in the multi-index which stores information
        about the unit of each timeseries.

    time_units
        Units of the time axis.

    Returns
    -------
    x_values :
        x-values (for a plot)

    y_values :
        y-values (for a plot)

    Raises
    ------
    TypeError
        `unit_aware` is not `False` and `unit_var` or `time_units` is `None`.

    MissingOptionalDependencyError
        `unit_aware` is `True`
        and [pint](https://pint.readthedocs.io/) is not installed.
    """
    res_no_units = (pdf.columns.values.squeeze(), pdf.values.squeeze())
    if not unit_aware:
        return res_no_units

    if unit_var is None:
        msg = "If `unit_aware` != False, then `unit_var` must not be `None`"
        raise TypeError(msg)

    if time_units is None:
        msg = "If `unit_aware` != False, then `time_units` must not be `None`"
        raise TypeError(msg)

    if isinstance(unit_aware, bool):
        try:
            import pint
        except ImportError as exc:
            raise MissingOptionalDependencyError(  # noqa: TRY003
                "get_values_line(..., unit_aware=True, ...)", requirement="pint"
            ) from exc

        ur = pint.get_application_registry()  # type: ignore

    else:
        ur = unit_aware

    res = (
        res_no_units[0] * ur(time_units),
        res_no_units[1] * ur(extract_single_unit(pdf, unit_var)),
    )

    return res


@overload
def get_values_plume(
    pdf: pd.DataFrame,
    *,
    quantiles: tuple[float, float],
    quantile_var: str,
    unit_aware: Literal[False],
    unit_var: str | None,
    time_units: str | None,
) -> tuple[
    NP_ARRAY_OF_FLOAT_OR_INT, NP_ARRAY_OF_FLOAT_OR_INT, NP_ARRAY_OF_FLOAT_OR_INT
]: ...


@overload
def get_values_plume(
    pdf: pd.DataFrame,
    *,
    quantiles: tuple[float, float],
    quantile_var: str,
    unit_aware: Literal[True] | pint.facets.PlainRegistry,
    unit_var: str | None,
    time_units: str | None,
) -> tuple[PINT_NUMPY_ARRAY, PINT_NUMPY_ARRAY, PINT_NUMPY_ARRAY]: ...


def get_values_plume(  # noqa: PLR0913
    pdf: pd.DataFrame,
    *,
    quantiles: tuple[float, float],
    quantile_var: str,
    unit_aware: bool | pint.facets.PlainRegistry,
    unit_var: str | None,
    time_units: str | None,
) -> (
    tuple[NP_ARRAY_OF_FLOAT_OR_INT, NP_ARRAY_OF_FLOAT_OR_INT, NP_ARRAY_OF_FLOAT_OR_INT]
    | tuple[PINT_NUMPY_ARRAY, PINT_NUMPY_ARRAY, PINT_NUMPY_ARRAY]
):
    """
    Get values for plotting a line

    Parameters
    ----------
    pdf
        [pd.DataFrame][pandas.DataFrame] from which to get the values

    quantiles
        Quantiles to get from `pdf`

    quantile_var
        Variable/column in the multi-index which stores information
        about the quantile that each timeseries represents.

    unit_aware
        Should the values be unit-aware?

        If `True`, we use the default application registry
        (retrieved with [pint.get_application_registry][]).
        Otherwise, a [pint.facets.PlainRegistry][] can be supplied and will be used.

    unit_var
        Variable/column in the multi-index which stores information
        about the unit of each timeseries.

    time_units
        Units of the time axis.

    Returns
    -------
    x_values :
        x-values (for a plot)

    y_values_lower :
        y-values for the lower-bound (of a plume plot)

    y_values_upper :
        y-values for the upper-bound (of a plume plot)

    Raises
    ------
    TypeError
        `unit_aware` is not `False` and `unit_var` or `time_units` is `None`.

    MissingOptionalDependencyError
        `unit_aware` is `True`
        and [pint](https://pint.readthedocs.io/) is not installed.
    """
    res_no_units = (
        pdf.columns.values.squeeze(),
        pdf.loc[
            pdf.index.get_level_values(quantile_var).isin({quantiles[0]})
        ].values.squeeze(),
        pdf.loc[
            pdf.index.get_level_values(quantile_var).isin({quantiles[1]})
        ].values.squeeze(),
    )
    if not unit_aware:
        return res_no_units

    if unit_var is None:
        msg = "If `unit_aware` != False, then `unit_var` must not be `None`"
        raise TypeError(msg)

    if time_units is None:
        msg = "If `unit_aware` != False, then `time_units` must not be `None`"
        raise TypeError(msg)

    if isinstance(unit_aware, bool):
        try:
            import pint
        except ImportError as exc:
            raise MissingOptionalDependencyError(  # noqa: TRY003
                "get_values_plume(..., unit_aware=True, ...)", requirement="pint"
            ) from exc

        ur = pint.get_application_registry()  # type: ignore

    else:
        ur = unit_aware

    unit = extract_single_unit(pdf, unit_var)
    res = (
        res_no_units[0] * ur(time_units),
        res_no_units[1] * ur(unit),
        res_no_units[2] * ur(unit),
    )

    return res


def create_legend_default(
    ax: matplotlib.axes.Axes, handles: list[matplotlib.artist.Artist]
) -> None:
    """
    Create legend, default implementation

    Intended to be used with [plot_plume_func][(m).]

    Parameters
    ----------
    ax
        Axes on which to create the legend

    handles
        Handles to include in the legend
    """
    ax.legend(handles=handles, loc="center left", bbox_to_anchor=(1.05, 0.5))


def get_default_colour_cycler() -> Iterator[COLOUR_VALUE_LIKE]:
    """
    Get the default colour cycler

    Returns
    -------
    :
        Default colour cycler

    Raises
    ------
    MissingOptionalDependencyError
        [matplotlib][] is not installed
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_default_colour_cycler", requirement="matplotlib"
        ) from exc

    colour_cycler = cycle(plt.rcParams["axes.prop_cycle"].by_key()["color"])

    return colour_cycler


def fill_out_palette(
    hue_values: Iterable[T],
    palette_user_supplied: PALETTE_LIKE[T] | None,
    warn_on_value_missing: bool,
) -> PALETTE_LIKE[T]:
    """
    Fill out a palette

    Parameters
    ----------
    hue_values
        Values which require a value in the output palette

    palette_user_supplied
        User-supplied palette

    warn_on_value_missing
        Should a warning be emitted if `palette_user_supplied` is not `None`
        but there are values missing from `palette_user_supplied`?

    Returns
    -------
    :
        Palette with values for all `hue_values`

    Warns
    -----
    UserWarning
        `warn_on_value_missing` is `True`,
        `palette_user_supplied` is not `None`
        and there are values in `hue_values` which are not in `palette_user_supplied`.
    """
    if palette_user_supplied is None:
        # Make it all ourselves.
        # Don't warn as the user didn't set any values
        # so it is clear they want us to fill in everything.
        colour_cycler = get_default_colour_cycler()
        palette_out: PALETTE_LIKE[T] = {  # type: ignore # not sure what I've done wrong
            v: next(colour_cycler) for v in hue_values
        }

        return palette_out

    # User-supplied palette
    missing_from_user_supplied = [
        v for v in hue_values if v not in palette_user_supplied
    ]
    if not missing_from_user_supplied:
        # Just return the values we need
        return {v: palette_user_supplied[v] for v in hue_values}  # type: ignore # not sure what mypy doesn't like

    if warn_on_value_missing:
        msg = (
            f"Some hue values are not in the user-supplied palette, "
            "they will be filled from the default colour cycler instead. "
            f"{missing_from_user_supplied=} {palette_user_supplied=}"
        )
        warnings.warn(msg)

    colour_cycler = get_default_colour_cycler()
    palette_out = {  # type: ignore # not sure what I've done wrong
        k: (
            palette_user_supplied[k]
            if k in palette_user_supplied
            else next(colour_cycler)
        )
        for k in hue_values
    }

    return palette_out


def get_default_dash_cycler() -> Iterator[DASH_VALUE_LIKE]:
    """
    Get the default dash cycler

    Returns
    -------
    :
        Default dash cycler
    """
    dash_cycler = cycle(["-", "--", "-.", ":"])

    return dash_cycler


def fill_out_dashes(
    style_values: Iterable[T],
    dashes_user_supplied: dict[T, DASH_VALUE_LIKE] | None,
    warn_on_value_missing: bool,
) -> dict[T, DASH_VALUE_LIKE]:
    """
    Fill out dashes

    Parameters
    ----------
    style_values
        Values which require a value in the output dashes

    dashes_user_supplied
        User-supplied dashes

    warn_on_value_missing
        Should a warning be emitted if `dashes_user_supplied` is not `None`
        but there are values missing from `dashes_user_supplied`?

    Returns
    -------
    :
        Dashes with values for all `style_values`

    Warns
    -----
    UserWarning
        `warn_on_value_missing` is `True`,
        `dashes_user_supplied` is not `None`
        and there are values in `style_values` which are not in `dashes_user_supplied`.
    """
    if dashes_user_supplied is None:
        # Make it all ourselves.
        # Don't warn as the user didn't set any values
        # so it is clear they want us to fill in everything.
        dash_cycler = get_default_dash_cycler()
        dashes_out = {v: next(dash_cycler) for v in style_values}

        return dashes_out

    # User-supplied palette
    missing_from_user_supplied = [
        v for v in style_values if v not in dashes_user_supplied
    ]
    if not missing_from_user_supplied:
        # Just return the values we need
        return {v: dashes_user_supplied[v] for v in style_values}

    if warn_on_value_missing:
        msg = (
            f"Some style values are not in the user-supplied dashes, "
            "they will be filled from the default dash cycler instead. "
            f"{missing_from_user_supplied=} {dashes_user_supplied=}"
        )
        warnings.warn(msg)

    dashes_out = {}
    dash_cycler = get_default_dash_cycler()
    for v in style_values:
        dashes_out[v] = (
            dashes_user_supplied[v] if v in dashes_user_supplied else next(dash_cycler)
        )

    return dashes_out


def same_shape_as_x_vals(
    obj: SingleLinePlotter | SinglePlumePlotter,
    attribute: attr.Attribute[Any],
    value: NP_ARRAY_OF_FLOAT_OR_INT | PINT_NUMPY_ARRAY,
) -> None:
    """
    Validate that the received values are the same shape as `obj.x_vals`

    Parameters
    ----------
    obj
        Object on which we are peforming validation

    attribute
        Attribute which is being set

    value
        Value which is being used to set `attribute`

    Raises
    ------
    AssertionError
        `value.shape` is not the same as `obj.x_vals.shape`
    """
    if value.shape != obj.x_vals.shape:
        msg = (
            f"`{attribute.name}` must have the same shape as `x_vals`. "
            f"Received `y_vals` with shape {value.shape} "
            f"while `x_vals` has shape {obj.x_vals.shape}"
        )
        raise AssertionError(msg)


@define
class SingleLinePlotter:
    """Object which is able to plot single lines"""

    x_vals: NP_ARRAY_OF_FLOAT_OR_INT | PINT_NUMPY_ARRAY
    """x-values to plot"""

    y_vals: NP_ARRAY_OF_FLOAT_OR_INT | PINT_NUMPY_ARRAY = field(
        validator=[same_shape_as_x_vals]
    )
    """y-values to plot"""

    quantile: float
    """Quantile that this line represents"""

    linewidth: float
    """Linewidth to use when plotting the line"""

    linestyle: DASH_VALUE_LIKE
    """Style to use when plotting the line"""

    color: COLOUR_VALUE_LIKE
    """Colour to use when plotting the line"""

    alpha: float
    """Alpha to use when plotting the line"""

    pkwargs: dict[str, Any] | None = None
    """Other arguments to pass to [matplotlib.axes.Axes.plot][] when plotting"""

    def get_label(self, quantile_legend_round: int = 2) -> str:
        """
        Get the label for the line

        Parameters
        ----------
        quantile_legend_round
            Rounding to apply to the quantile when creating the label

        Returns
        -------
        :
            Label for the line
        """
        label = str(np.round(self.quantile, quantile_legend_round))

        return label

    def plot(self, ax: matplotlib.axes.Axes, quantile_legend_round: int = 2) -> None:
        """
        Plot

        Parameters
        ----------
        ax
            Axes on which to plot

        quantile_legend_round
            Rounding to apply to the quantile when creating the label
        """
        pkwargs = self.pkwargs if self.pkwargs is not None else {}

        ax.plot(
            self.x_vals,
            self.y_vals,
            label=self.get_label(quantile_legend_round=quantile_legend_round),
            linewidth=self.linewidth,
            linestyle=self.linestyle,
            color=self.color,
            alpha=self.alpha,
            **pkwargs,
        )


@define
class SinglePlumePlotter:
    """Object which is able to plot single plumes"""

    x_vals: NP_ARRAY_OF_FLOAT_OR_INT | PINT_NUMPY_ARRAY
    """x-values to plot"""

    y_vals_lower: NP_ARRAY_OF_FLOAT_OR_INT | PINT_NUMPY_ARRAY = field(
        validator=[same_shape_as_x_vals]
    )
    """y-values to plot as the lower bound of the plume"""

    y_vals_upper: NP_ARRAY_OF_FLOAT_OR_INT | PINT_NUMPY_ARRAY = field(
        validator=[same_shape_as_x_vals]
    )
    """y-values to plot as the upper bound of the plume"""

    quantiles: tuple[float, float]
    """Quantiles that this plume represents"""

    color: COLOUR_VALUE_LIKE
    """Colour to use when plotting the plume"""

    alpha: float
    """Alpha to use when plotting the plume"""

    pkwargs: dict[str, Any] | None = None
    """Other arguments to pass to [matplotlib.axes.Axes.fill_between][] when plotting"""

    def get_label(self, quantile_legend_round: int = 2) -> str:
        """
        Get the label for the plume

        Parameters
        ----------
        quantile_legend_round
            Rounding to apply to the quantiles when creating the label

        Returns
        -------
        :
            Label for the plume
        """
        label = " - ".join(
            [str(np.round(qv, quantile_legend_round)) for qv in self.quantiles]
        )

        return label

    def plot(self, ax: matplotlib.axes.Axes, quantile_legend_round: int = 2) -> None:
        """
        Plot

        Parameters
        ----------
        ax
            Axes on which to plot

        quantile_legend_round
            Rounding to apply to the quantiles when creating the label
        """
        pkwargs = self.pkwargs if self.pkwargs is not None else {}

        ax.fill_between(
            self.x_vals,
            self.y_vals_lower,
            self.y_vals_upper,
            label=self.get_label(quantile_legend_round=quantile_legend_round),
            facecolor=self.color,
            alpha=self.alpha,
            **pkwargs,
        )


@define
class PlumePlotter:
    """Object which is able to plot plume plots"""

    lines: Iterable[SingleLinePlotter]
    """Lines to plot"""

    plumes: Iterable[SinglePlumePlotter]
    """Lines to plot"""

    hue_var_label: str
    """Label for the hue variable in the legend"""

    style_var_label: str | None
    """Label for the style variable in the legend (if not `None`)"""

    quantile_var_label: str
    """Label for the quantile variable in the legend"""

    palette: PALETTE_LIKE[Any]
    """Palette used for plotting different values of the hue variable"""

    dashes: dict[Any, str | tuple[float, tuple[float, ...]]] | None
    """Dashes used for plotting different values of the style variable"""

    x_label: str | None
    """Label to apply to the x-axis (if `None`, no label is applied)"""

    y_label: str | None
    """Label to apply to the y-axis (if `None`, no label is applied)"""

    @classmethod
    def from_df(  # noqa: PLR0912, PLR0913, PLR0915 # object creation code is the worst
        cls,
        df: pd.DataFrame,
        *,
        quantiles_plumes: QUANTILES_PLUMES_LIKE = (
            (0.5, 0.7),
            ((0.05, 0.95), 0.2),
        ),
        quantile_var: str = "quantile",
        quantile_var_label: str | None = None,
        quantile_legend_round: int = 2,
        hue_var: str = "scenario",
        hue_var_label: str | None = None,
        palette: PALETTE_LIKE[Any] | None = None,
        warn_on_palette_value_missing: bool = True,
        style_var: str | None = "variable",
        style_var_label: str | None = None,
        dashes: dict[Any, str | tuple[float, tuple[float, ...]]] | None = None,
        warn_on_dashes_value_missing: bool = True,
        linewidth: float = 3.0,
        unit_var: str | None = "unit",
        unit_aware: bool | pint.facets.PlainRegistry = False,
        time_units: str | None = None,
        x_label: str | None = "time",
        y_label: str | bool | None = True,
        warn_infer_y_label_with_multi_unit: bool = True,
        observed: bool = True,
    ) -> PlumePlotter:
        """
        Initialise from a [pd.DataFrame][pandas.DataFrame]

        Parameters
        ----------
        df
            [pd.DataFrame][pandas.DataFrame] from which to initialise

        quantiles_plumes
            Quantiles to plot in each plume.

            If the first element of each tuple is a tuple,
            a [SinglePlumePlotter][(m).] object will be created.
            Otherwise, if the first element is a plain float,
            a [SingleLinePlotter][(m).] object will be created.

        quantile_var
            Variable/column in the multi-index which stores information
            about the quantile that each timeseries represents.

        quantile_var_label
            Label to use as the header for the quantile section in the legend

        quantile_legend_round
            Rounding to apply to quantile values when creating the legend

        hue_var
            Variable to use for grouping data into different colour groups

        hue_var_label
            Label to use as the header for the hue/colour section in the legend

        palette
            Colour to use for the different groups in the data.

            If any groups are not included in `palette`,
            they are auto-filled.

        warn_on_palette_value_missing
            Should a warning be emitted if there are values missing from `palette`?

        style_var
            Variable to use for grouping data into different (line)style groups

        style_var_label
            Label to use as the header for the style section in the legend

        dashes
            Dash/linestyle to use for the different groups in the data.

            If any groups are not included in `dashes`,
            they are auto-filled.

        warn_on_dashes_value_missing
            Should a warning be emitted if there are values missing from `dashes`?

        linewidth
            Width to use for plotting lines.

        unit_var
            Variable/column in the multi-index which stores information
            about the unit of each timeseries.

        unit_aware
            Should the values be extracted in a unit-aware way?

            If `True`, we use the default application registry
            (retrieved with [pint.get_application_registry][]).
            Otherwise, a [pint.facets.PlainRegistry][] can be supplied and will be used.

        time_units
            Units of the time axis of the data.

            These are required if `unit_aware` is not `False`.

        x_label
            Label to apply to the x-axis.

            If `None`, no label will be applied.

        y_label
            Label to apply to the y-axis.

            If `True`, we will try and infer the y-label based on the data's units.

            If `None`, no label will be applied.

        warn_infer_y_label_with_multi_unit
            Should a warning be raised if we try to infer the y-unit
            but the data has more than one unit?

        observed
            Passed to [pd.DataFrame.groupby][pandas.DataFrame.groupby].

        Returns
        -------
        :
             Initialised instance
        """
        if hue_var_label is None:
            hue_var_label = hue_var.capitalize()

        if style_var is not None and style_var_label is None:
            style_var_label = style_var.capitalize()

        if quantile_var_label is None:
            quantile_var_label = quantile_var.capitalize()

        infer_y_label = (
            (not unit_aware)
            and isinstance(y_label, bool)
            and y_label
            and unit_var is not None
        )

        palette_complete = fill_out_palette(
            df.index.get_level_values(hue_var).unique(),
            palette_user_supplied=palette,
            warn_on_value_missing=warn_on_palette_value_missing,
        )

        if style_var is not None:
            group_cols = [hue_var, style_var]
            dashes_complete = fill_out_dashes(
                df.index.get_level_values(style_var).unique(),
                dashes_user_supplied=dashes,
                warn_on_value_missing=warn_on_dashes_value_missing,
            )

        else:
            group_cols = [hue_var]
            dashes_complete = None

        lines: list[SingleLinePlotter] = []
        plumes: list[SinglePlumePlotter] = []
        values_units: list[str] = []
        for info, gdf in df.groupby(group_cols, observed=observed):
            info_d = {k: v for k, v in zip(group_cols, info)}

            colour = palette_complete[info_d[hue_var]]

            gpdf = partial(get_pdf_from_pre_calculated, gdf, quantile_col=quantile_var)

            def warn_about_missing_quantile(exc: Exception) -> None:
                warnings.warn(
                    f"Quantiles missing for {info_d}. Original exception: {exc}"
                )

            for q, alpha in quantiles_plumes:
                if isinstance(q, float):
                    if style_var is not None:
                        if dashes_complete is None:  # pragma: no cover
                            # should be impossible to hit this
                            raise AssertionError
                        linestyle = dashes_complete[info_d[style_var]]
                    else:
                        linestyle = "-"

                    try:
                        quantiles = (q,)
                        pdf = gpdf(quantiles=quantiles)
                    except MissingQuantileError as exc:
                        warn_about_missing_quantile(exc=exc)
                        continue

                    line_plotter = SingleLinePlotter(
                        *get_values_line(
                            pdf,
                            unit_aware=unit_aware,  # type: ignore # not sure why mypy is complaining
                            unit_var=unit_var,
                            time_units=time_units,
                        ),
                        quantile=q,
                        linewidth=linewidth,
                        linestyle=linestyle,
                        color=colour,
                        alpha=alpha,
                    )
                    lines.append(line_plotter)

                else:
                    try:
                        pdf = gpdf(quantiles=q)
                    except MissingQuantileError as exc:
                        warn_about_missing_quantile(exc=exc)
                        continue

                    plume_plotter = SinglePlumePlotter(
                        *get_values_plume(
                            pdf,
                            quantiles=q,
                            quantile_var=quantile_var,
                            unit_aware=unit_aware,  # type: ignore # not sure why mypy is complaining
                            unit_var=unit_var,
                            time_units=time_units,
                        ),
                        quantiles=q,
                        color=colour,
                        alpha=alpha,
                    )
                    plumes.append(plume_plotter)

                if infer_y_label and unit_var in pdf.index.names:
                    values_units.extend(pdf.index.get_level_values(unit_var).unique())

        if unit_aware and isinstance(y_label, bool) and y_label:
            # Let unit-aware plotting do its thing
            y_label = None

        elif unit_var is None:
            y_label = None

        elif infer_y_label:
            if unit_var not in df.index.names:
                warnings.warn(
                    "Not auto-setting the y_label "
                    f"because {unit_var=} is not in {df.index.names=}"
                )
                y_label = None

            else:
                # Try to infer the y-label
                units_s = set(values_units)
                if len(units_s) == 1:
                    y_label = values_units[0]
                else:
                    # More than one unit plotted, don't infer a y-label
                    if warn_infer_y_label_with_multi_unit:
                        warnings.warn(
                            "Not auto-setting the y_label "
                            "because the plotted data has more than one unit: "
                            f"data units {units_s}"
                        )

                    y_label = None

        if isinstance(y_label, bool):
            msg = "y_label should have been converted before getting here"
            raise TypeError(msg)

        res = PlumePlotter(
            lines=lines,
            plumes=plumes,
            hue_var_label=hue_var_label,
            style_var_label=style_var_label,
            quantile_var_label=quantile_var_label,
            palette=palette_complete,
            dashes=dashes_complete,
            x_label=x_label,
            y_label=y_label,
        )

        return res

    def generate_legend_handles(
        self, quantile_legend_round: int = 2
    ) -> list[matplotlib.artist.Artist]:
        """
        Generate handles for the legend

        Parameters
        ----------
        quantile_legend_round
            Rounding to apply to the quantiles when creating the label

        Returns
        -------
        :
            Generated handles for the legend
        """
        try:
            import matplotlib.lines as mlines
            import matplotlib.patches as mpatches
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "generate_legend_handles", requirement="matplotlib"
            ) from exc

        generated_quantile_items: list[
            Union[
                tuple[float, float, str],
                tuple[tuple[float, float], float, str],
            ]
        ] = []
        quantile_items: list[matplotlib.artist.Artist] = []
        for line in self.lines:
            label = line.get_label(quantile_legend_round=quantile_legend_round)
            pid_line = (line.quantile, line.alpha, label)
            if pid_line in generated_quantile_items:
                continue

            quantile_items.append(
                mlines.Line2D([0], [0], color="k", alpha=line.alpha, label=label)
            )
            generated_quantile_items.append(pid_line)

        for plume in self.plumes:
            label = plume.get_label(quantile_legend_round=quantile_legend_round)
            pid_plume = (plume.quantiles, plume.alpha, label)
            if pid_plume in generated_quantile_items:
                continue

            quantile_items.append(
                mpatches.Patch(color="k", alpha=plume.alpha, label=label)
            )
            generated_quantile_items.append(pid_plume)

        hue_items = [
            mlines.Line2D([0], [0], color=colour, label=hue_value)
            for hue_value, colour in self.palette.items()
        ]

        legend_items = [
            mpatches.Patch(alpha=0, label=self.quantile_var_label),
            *quantile_items,
            mpatches.Patch(alpha=0, label=self.hue_var_label),
            *hue_items,
        ]
        if self.dashes is not None and self.lines:
            style_items = [
                mlines.Line2D(
                    [0],
                    [0],
                    linestyle=linestyle,
                    label=style_value,
                    color="gray",
                )
                for style_value, linestyle in self.dashes.items()
            ]
            legend_items.append(mpatches.Patch(alpha=0, label=self.style_var_label))
            legend_items.extend(style_items)

        return legend_items

    def plot(
        self,
        ax: matplotlib.axes.Axes | None = None,
        *,
        create_legend: Callable[
            [matplotlib.axes.Axes, list[matplotlib.artist.Artist]], None
        ] = create_legend_default,
        quantile_legend_round: int = 2,
    ) -> matplotlib.axes.Axes:
        """
        Plot

        Parameters
        ----------
        ax
            Axes onto which to plot

        create_legend
            Function to use to create the legend.

            This allows the user to have full control over the creation of the legend.

        quantile_legend_round
            Rounding to apply to quantile values when creating the legend

        Returns
        -------
        :
            Axes on which the data was plotted
        """
        if ax is None:
            try:
                import matplotlib.pyplot as plt
            except ImportError as exc:
                raise MissingOptionalDependencyError(  # noqa: TRY003
                    "plot(ax=None, ...)", requirement="matplotlib"
                ) from exc

            _, ax = plt.subplots()

        for plume in self.plumes:
            plume.plot(ax=ax)

        for line in self.lines:
            line.plot(ax=ax)

        create_legend(
            ax,
            self.generate_legend_handles(quantile_legend_round=quantile_legend_round),
        )

        if self.x_label is not None:
            ax.set_xlabel(self.x_label)

        if self.y_label is not None:
            ax.set_ylabel(self.y_label)

        return ax


# Something funny happening with relative x-refs, hence _func suffix
def plot_plume_func(  # noqa: PLR0913
    pdf: pd.DataFrame,
    quantiles_plumes: QUANTILES_PLUMES_LIKE,
    ax: matplotlib.axes.Axes | None = None,
    *,
    quantile_var: str = "quantile",
    quantile_var_label: str | None = None,
    quantile_legend_round: int = 3,
    hue_var: str = "scenario",
    hue_var_label: str | None = None,
    palette: PALETTE_LIKE[Any] | None = None,
    warn_on_palette_value_missing: bool = True,
    style_var: str = "variable",
    style_var_label: str | None = None,
    dashes: dict[Any, str | tuple[float, tuple[float, ...]]] | None = None,
    warn_on_dashes_value_missing: bool = True,
    linewidth: float = 2.0,
    unit_var: str = "unit",
    unit_aware: bool | pint.facets.PlainRegistry = False,
    time_units: str | None = None,
    x_label: str | None = "time",
    y_label: str | bool | None = True,
    warn_infer_y_label_with_multi_unit: bool = True,
    create_legend: Callable[
        [matplotlib.axes.Axes, list[matplotlib.artist.Artist]], None
    ] = create_legend_default,
    observed: bool = True,
) -> matplotlib.axes.Axes:
    """
    Plot a plume plot

    Parameters
    ----------
    pdf
        [pd.DataFrame][pandas.DataFrame] to use for plotting

        It must contain quantiles already.
        For data without quantiles, please see
        [plot_plume_after_calculating_quantiles_func][(m).].

    quantiles_plumes
        Quantiles to plot in each plume.

        If the first element of each tuple is a tuple,
        a plume is plotted between the given quantiles.
        Otherwise, if the first element is a plain float,
        a line is plotted for the given quantile.

    ax
        Axes on which to plot.

        If not supplied, a new axes is created.

    quantile_var
        Variable/column in the multi-index which stores information
        about the quantile that each timeseries represents.

    quantile_var_label
        Label to use as the header for the quantile section in the legend

    quantile_legend_round
        Rounding to apply to quantile values when creating the legend

    hue_var
        Variable to use for grouping data into different colour groups

    hue_var_label
        Label to use as the header for the hue/colour section in the legend

    palette
        Colour to use for the different groups in the data.

        If any groups are not included in `palette`,
        they are auto-filled.

    warn_on_palette_value_missing
        Should a warning be emitted if there are values missing from `palette`?

    style_var
        Variable to use for grouping data into different (line)style groups

    style_var_label
        Label to use as the header for the style section in the legend

    dashes
        Dash/linestyle to use for the different groups in the data.

        If any groups are not included in `dashes`,
        they are auto-filled.

    warn_on_dashes_value_missing
        Should a warning be emitted if there are values missing from `dashes`?

    linewidth
        Width to use for plotting lines.

    unit_var
        Variable/column in the multi-index which stores information
        about the unit of each timeseries.

    unit_aware
        Should the plot be done in a unit-aware way?

        If `True`, we use the default application registry
        (retrieved with [pint.get_application_registry][]).
        Otherwise, a [pint.facets.PlainRegistry][] can be supplied and will be used.

        For details, see matplotlib and pint support plotting with units
        ([stable docs](https://pint.readthedocs.io/en/stable/user/plotting.html),
        [last version that we checked at the time of writing](https://pint.readthedocs.io/en/0.24.4/user/plotting.html)).

    time_units
        Units of the time axis of the data.

        These are required if `unit_aware` is not `False`.

    x_label
        Label to apply to the x-axis.

        If `None`, no label will be applied.

    y_label
        Label to apply to the y-axis.

        If `True`, we will try and infer the y-label based on the data's units.

        If `None`, no label will be applied.

    warn_infer_y_label_with_multi_unit
        Should a warning be raised if we try to infer the y-unit
        but the data has more than one unit?

    create_legend
        Function to use to create the legend.

        This allows the user to have full control over the creation of the legend.

    observed
        Passed to [pd.DataFrame.groupby][pandas.DataFrame.groupby].

    Returns
    -------
    :
        Axes on which the data was plotted
    """
    plotter = PlumePlotter.from_df(
        df=pdf,
        quantiles_plumes=quantiles_plumes,
        quantile_var=quantile_var,
        quantile_var_label=quantile_var_label,
        hue_var=hue_var,
        hue_var_label=hue_var_label,
        palette=palette,
        warn_on_palette_value_missing=warn_on_palette_value_missing,
        style_var=style_var,
        style_var_label=style_var_label,
        dashes=dashes,
        warn_on_dashes_value_missing=warn_on_dashes_value_missing,
        linewidth=linewidth,
        unit_var=unit_var,
        unit_aware=unit_aware,
        time_units=time_units,
        x_label=x_label,
        y_label=y_label,
        warn_infer_y_label_with_multi_unit=warn_infer_y_label_with_multi_unit,
        observed=observed,
    )

    ax = plotter.plot(
        ax=ax, create_legend=create_legend, quantile_legend_round=quantile_legend_round
    )

    return ax


# Something funny happening with relative x-refs, hence _func suffix
def plot_plume_after_calculating_quantiles_func(  # noqa: PLR0913
    pdf: pd.DataFrame,
    ax: matplotlib.axes.Axes | None = None,
    *,
    quantile_over: str | list[str],
    quantiles_plumes: QUANTILES_PLUMES_LIKE = (
        (0.5, 0.7),
        ((0.05, 0.95), 0.2),
    ),
    quantile_var_label: str | None = None,
    quantile_legend_round: int = 2,
    hue_var: str = "scenario",
    hue_var_label: str | None = None,
    palette: PALETTE_LIKE[Any] | None = None,
    warn_on_palette_value_missing: bool = True,
    style_var: str = "variable",
    style_var_label: str | None = None,
    dashes: dict[Any, str | tuple[float, tuple[float, ...]]] | None = None,
    warn_on_dashes_value_missing: bool = True,
    linewidth: float = 3.0,
    unit_var: str = "unit",
    unit_aware: bool | pint.facets.PlainRegistry = False,
    time_units: str | None = None,
    x_label: str | None = "time",
    y_label: str | bool | None = True,
    warn_infer_y_label_with_multi_unit: bool = True,
    create_legend: Callable[
        [matplotlib.axes.Axes, list[matplotlib.artist.Artist]], None
    ] = create_legend_default,
    observed: bool = True,
) -> matplotlib.axes.Axes:
    """
    Plot a plume plot, calculating the required quantiles first

    Parameters
    ----------
    pdf
        [pd.DataFrame][pandas.DataFrame] to use for plotting

        It must contain quantiles already.
        For data without quantiles, please see
        [plot_plume_after_calculating_quantiles_func][(m).].

    ax
        Axes on which to plot.

        If not supplied, a new axes is created.

    quantile_over
        Variable(s)/column(s) over which to calculate the quantiles.

        The data is grouped by all columns except `quantile_over`
        when calculating the quantiles.

    quantiles_plumes
        Quantiles to plot in each plume.

        If the first element of each tuple is a tuple,
        a plume is plotted between the given quantiles.
        Otherwise, if the first element is a plain float,
        a line is plotted for the given quantile.

    quantile_var_label
        Label to use as the header for the quantile section in the legend

    quantile_legend_round
        Rounding to apply to quantile values when creating the legend

    hue_var
        Variable to use for grouping data into different colour groups

    hue_var_label
        Label to use as the header for the hue/colour section in the legend

    palette
        Colour to use for the different groups in the data.

        If any groups are not included in `palette`,
        they are auto-filled.

    warn_on_palette_value_missing
        Should a warning be emitted if there are values missing from `palette`?

    style_var
        Variable to use for grouping data into different (line)style groups

    style_var_label
        Label to use as the header for the style section in the legend

    dashes
        Dash/linestyle to use for the different groups in the data.

        If any groups are not included in `dashes`,
        they are auto-filled.

    warn_on_dashes_value_missing
        Should a warning be emitted if there are values missing from `dashes`?

    linewidth
        Width to use for plotting lines.

    unit_var
        Variable/column in the multi-index which stores information
        about the unit of each timeseries.

    unit_aware
        Should the plot be done in a unit-aware way?

        If `True`, we use the default application registry
        (retrieved with [pint.get_application_registry][]).
        Otherwise, a [pint.facets.PlainRegistry][] can be supplied and will be used.

        For details, see matplotlib and pint support plotting with units
        ([stable docs](https://pint.readthedocs.io/en/stable/user/plotting.html),
        [last version that we checked at the time of writing](https://pint.readthedocs.io/en/0.24.4/user/plotting.html)).

    time_units
        Units of the time axis.

        These are required if `unit_aware` is not `False`.

    x_label
        Label to apply to the x-axis.

        If `None`, no label will be applied.

    y_label
        Label to apply to the y-axis.

        If `True`, we will try and infer the y-label based on the data's units.

        If `None`, no label will be applied.

    warn_infer_y_label_with_multi_unit
        Should a warning be raised if we try to infer the y-unit
        but the data has more than one unit?

    create_legend
        Function to use to create the legend.

        This allows the user to have full control over the creation of the legend.

    observed
        Passed to [pd.DataFrame.groupby][pandas.DataFrame.groupby].

    Returns
    -------
    :
        Axes on which the data was plotted
    """
    quantile_var = "quantile"
    pdf_q = fix_index_name_after_groupby_quantile(
        groupby_except(pdf, quantile_over).quantile(get_quantiles(quantiles_plumes)),
        new_name=quantile_var,
        copy=False,
    )

    return plot_plume_func(
        pdf=pdf_q,
        ax=ax,
        quantiles_plumes=quantiles_plumes,
        quantile_var=quantile_var,
        quantile_var_label=quantile_var_label,
        quantile_legend_round=quantile_legend_round,
        hue_var=hue_var,
        hue_var_label=hue_var_label,
        palette=palette,
        warn_on_palette_value_missing=warn_on_palette_value_missing,
        style_var=style_var,
        style_var_label=style_var_label,
        dashes=dashes,
        warn_on_dashes_value_missing=warn_on_dashes_value_missing,
        linewidth=linewidth,
        unit_var=unit_var,
        unit_aware=unit_aware,
        time_units=time_units,
        x_label=x_label,
        y_label=y_label,
        warn_infer_y_label_with_multi_unit=warn_infer_y_label_with_multi_unit,
        create_legend=create_legend,
        observed=observed,
    )
